"""Function derivatives for error propagation."""

import sys
import numpy as np
import copy


__all__ = ['derivatives', 'propagate_1', 'propagate_2']


STEP_SIZE = np.sqrt(sys.float_info.epsilon)


@np.vectorize
def _deriv_pow_0(x, y):
    """Partial derivative of x**y in x."""
    if y == 0:
        return 0.0
    if x != 0 or y % 1 == 0:
        return y*x**(y-1)
    return numerical_derivative(np.power, 0)(x, y)


@np.vectorize
def _deriv_pow_1(x, y):
    """Partial derivative of x**y in y."""
    if x == 0 and y > 0:
        return 0.0
    return np.log(x)*np.power(x, y)


@np.vectorize
def _deriv_mod_1(x, y):
    """Partial derivative of x%y in y."""
    if x % y < STEP_SIZE:
        return np.inf
    else:
        return numerical_derivative(np.mod, 1)(x, y)


@np.vectorize
def _deriv_fabs(x):
    if x >= 0:
        return 1
    return -1


@np.vectorize
def _deriv_copysign(x, y):
    if x >= 0:
        return np.copysign(1, y)
    return -np.copysign(1, y)


erf_coef = 2/np.sqrt(np.pi)


# Function partial derivatives for x and y
derivatives = {
    'add': (lambda x, y: 1.0,
            lambda x, y: 1.0),
    'sub': (lambda x, y: 1.0,
            lambda x, y: -1.0),
    'div': (lambda x, y: 1/y,
            lambda x, y: -x/(y**2)),
    'truediv': (lambda x, y: 1/y,
                lambda x, y: -x/(y**2)),
    'floordiv': (lambda x, y: 0.0,
                 lambda x, y: 0.0),
    'mod': (lambda x, y: 1.0,
            _deriv_mod_1),
    'mul': (lambda x, y: y,
            lambda x, y: x),
    'pow': (_deriv_pow_0, _deriv_pow_1),
    # numpy fixed derivatives
    'arccos': (lambda x: -1/np.sqrt(1-x**2)),
    'arccosh': (lambda x: 1/np.sqrt(x**2-1)),
    'arcsin': (lambda x: 1/np.sqrt(1-x**2)),
    'arcsinh': (lambda x: 1/np.sqrt(1+x**2)),
    'arctan': (lambda x: 1/(1+x**2)),
    'arctan2': (lambda y, x: x/(x**2+y**2),  # Correct for x == 0
                lambda y, x: -y/(x**2+y**2)),  # Correct for x == 0
    'arctanh': (lambda x: 1/(1-x**2)),
    'copysign': (_deriv_copysign,
                 lambda x, y: 0),
    'cos': (lambda x: -np.sin(x)),
    'cosh': (np.sinh),
    'exp': (np.exp),
    'expm1': (np.exp),
    'exp2': (lambda x: np.exp2(x)*np.log(2)),
    'fabs': (_deriv_fabs),
    'hypot': [lambda x, y: x/np.hypot(x, y),
              lambda x, y: y/np.hypot(x, y)],
    'log': (lambda x: 1/x),  # for np, log=ln
    'log10': (lambda x: 1/(x*np.log(10.0))),
    'log2': (lambda x: 1/(x*np.log(2.0))),
    'log1p': (lambda x: 1/(1+x)),
    'sin': (np.cos),
    'sinh': (np.cosh),
    'tan': (lambda x: 1+np.tan(x)**2),
    'tanh': (lambda x: 1-np.tanh(x)**2)
}


def propagate_1(func, fx, x, sx):
    """Propagate errors using function derivatives.

    Parameters
    ----------
    func: string
        Function name to perform error propagation. Must be in derivatives
        keys.
    fx: float or array_like
        Numerical result of f(x, y).
    x: float or array_like
        Variable of the function.
    sx: float or array_like
        1-sigma errors of the function variable.

    Returns
    -------
    sf: float or array_like
        1-sigma uncorrelated error associated to the operation.
    """
    if func not in derivatives.keys():
        raise ValueError(f'func {func} not in derivatives.')

    if hasattr(derivatives[func], '__len__'):
        raise ValueError(f'func {func} is not a 1 variable function.')

    try:
        deriv = derivatives[func](x)
        sf = deriv*sx
        return sf
    except (ValueError, ZeroDivisionError, OverflowError):
        shape = np.shape(fx)
        if len(shape) == 0:
            return np.nan
        else:
            return np.empty(shape).fill(np.nan)


def propagate_2(func, fxy, x, y, sx, sy):
    """Propagate errors using function derivatives.

    Parameters
    ----------
    func: string
        Function name to perform error propagation. Must be in derivatives
        keys.
    fxy: float or array_like
        Numerical result of f(x, y).
    x, y: float or array_like
        Variables of the function.
    sx, sy: float or array_like
        1-sigma errors of the function variables.

    Returns
    -------
    sf: float or array_like
        1-sigma uncorrelated error associated to the operation.
    """
    if func not in derivatives.keys():
        raise ValueError(f'func {func} not in derivatives.')

    if len(derivatives[func]) != 2:
        raise ValueError(f'func {func} is not a 2 variable function.')

    deriv_x, deriv_y = derivatives[func]
    try:
        del_x2 = np.square(deriv_x(x, y))
        del_y2 = np.square(deriv_y(x, y))
        sx2 = np.square(sx)
        sy2 = np.square(sy)
        sf = np.sqrt(del_x2*sx2 + del_y2*sy2)
        return sf
    except (ValueError, ZeroDivisionError, OverflowError):
        shape = np.shape(fxy)
        if len(shape) == 0:
            return np.nan
        else:
            return np.empty(shape).fill(np.nan)


def numerical_derivative(func, arg_ref, step=STEP_SIZE):
    """Create a function to compute a numerical derivative of func.

    Parameters
    ----------
    func: callable
        The function to compute the numerical derivative.
    arg_ref: int or string
        Variable to be used for diferentiation. If int, a position will be
        used. If string, a variable name will be used.
    step: float (optional)
        Epsilon to compute the numerical derivative, using the
        (-epsilon, +epsioln) method.

    Returns
    -------
    derivative_wrapper: callable
        Partial derivative function.

    Notes
    -----
    - Implementation based on `uncertainties` package.
    """
    if not callable(func):
        raise ValueError(f'function {func} not callable.')

    # If reference variable is in kwargs (str) instead of args (int).
    change_kwargs = isinstance(arg_ref, str)

    @np.vectorize
    def derivative_wrapper(*args, **kwargs):
        """
        Partial derivative, calculated with the (-epsilon, +epsilon)
        method, which is more precise than the (0, +epsilon) method.
        """
        # Compute the epsilon relative to the variyng parameter
        if change_kwargs:
            var_v = kwargs[arg_ref]
        else:
            var_v = args[arg_ref]

        eps = step*abs(var_v)
        if not eps:
            # Arbitrary, but "small" with respect to 1:
            eps = step

        # Copy args and assing the variable function with values plus
        # and minus the epsilon
        # This will cause more memory to be used, but do not change the
        # original objects.
        kwargs_p = copy.copy(kwargs)
        kwargs_m = copy.copy(kwargs)
        args_p = list(args)
        args_m = list(args)

        if change_kwargs:
            kwargs_p[arg_ref] += eps
            kwargs_m[arg_ref] -= eps
        else:
            args_p[arg_ref] += eps
            args_m[arg_ref] -= eps

        # Compute the function values with shifted vvariable
        f_plus = func(*args_p, **kwargs_p)
        f_minus = func(*args_m, **kwargs_m)

        return (f_plus - f_minus)/2/eps

    return derivative_wrapper



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Compute shifts and translate astronomical images."""

import abc
from functools import partial
from skimage import transform
import numpy as np

from .processing import trim_image
from ..logger import logger
from ..framedata import check_framedata, FrameData


__all__ = ['CrossCorrelationRegister', 'AsterismRegister',
           'register_framedata_list', 'compute_shift_list']


def _algorithm_check(algorithm, kwargs):
    """Check the algorithms and get the register."""
    if algorithm == 'cross-correlation':
        return CrossCorrelationRegister(**kwargs)
    if algorithm == 'asterism-matching':
        return AsterismRegister(**kwargs)
    raise ValueError(f'Algorithm {algorithm} unknown.')


def _check_compatible_list(frame_list):
    """Check if the list if it is compatible with the register."""
    for i in frame_list:
        if not isinstance(i, FrameData):
            raise TypeError('Only a list of FrameData instances is allowed.')
        if i.shape != frame_list[0].shape:
            raise ValueError('Images with incompatible shapes '
                             f'{frame_list[0].shape} and {i.shape}.'
                             'Only frames with same shape allowed.')


def _get_clip_slices(shifts, imshape):
    """Get the slices of the clip."""
    # negative shifts means the image is translated up and right, so they
    # define the start.
    # positive shifts translates down and left, so they define the stop.
    xmin, xmax = np.nanmin(shifts[:, 0]), np.nanmax(shifts[:, 0])
    ymin, ymax = np.nanmin(shifts[:, 1]), np.nanmax(shifts[:, 1])

    xstart = int(np.ceil(-xmin)) if xmin < 0 else 0
    xstop = int(np.floor(-xmax)) if xmax > 0 else imshape[0]

    ystart = int(np.ceil(-ymin)) if ymin < 0 else 0
    ystop = int(np.floor(-ymax)) if ymax > 0 else imshape[1]

    return slice(xstart, xstop), slice(ystart, ystop)


_keywords = {
    'method': 'HIERARCH astropop registration',
    'shift_x': 'HIERARCH astropop registration_shift_x',
    'shift_y': 'HIERARCH astropop registration_shift_y',
    'rotation': 'HIERARCH astropop registration_rot',
}


class _BaseRegister(abc.ABC):
    """Base class for Registers."""

    _name = None

    @staticmethod
    def _apply_transform_image(image, tform, cval=0):
        """Apply the transform to an image."""
        return transform.warp(image, tform, mode='constant', cval=cval,
                              preserve_range=True)

    @abc.abstractmethod
    def _compute_transform(self, image1, image2, mask1=None, mask2=None):
        """Compute the transform to register image2 to image1."""

    def compute_transform(self, image1, image2, mask1=None, mask2=None):
        """Compute the transform to register image2 to image1.

        Parameters
        ----------
        image1 : 2d `~numpy.ndarray`
            Reference image for registration.
        image2 : 2d `~numpy.ndarray`
            Moving image for registration.
        mask1, mask2 : 2d `~numpy.ndarray` (optional)
            Masks for images 1 and 2.
            Default: `None`

        Returns
        -------
        tfrom : `~skimage.transform.AffineTransform`
            Transform computed to project image2 in image2.
        """
        return self._compute_transform(image1, image2, mask1, mask2)

    def register_image(self, image1, image2, mask1=None, mask2=None,
                       cval='median'):
        """Align and transform an image2 to match image1.

        Parameters
        ----------
        image1 : 2d `~numpy.ndarray`
            Reference image for registration.
        image2 : 2d `~numpy.ndarray`
            Moving image for registration.
        mask1, mask2 : 2d `~numpy.ndarray` (optional)
            Masks for images 1 and 2.
            Default: `None`
        cval : float or {'median', 'mean'} (optional)
            Fill value for transformed pixels from outside the image. If
            'median' or 'mean', these statistics will be computed as cval.
            Default: 'median'

        Returns
        -------
        reg_image : `~numpy.ndarray`
            Registered image according the transform computed by the class.
        mask : `~numpy.ndarray`
            Mask for the registered image.
        tfrom : `~skimage.transform.AffineTransform`
            Transform computed to project image2 in image2.
        """
        # equal images are just returned
        if np.all(image1 == image2):
            logger.info('Images are equal, skipping registering.')
            return image1, np.zeros_like(image1), transform.AffineTransform()

        tform = self._compute_transform(image1, image2, mask1, mask2)
        if mask2 is None:
            mask2 = np.zeros_like(image2)

        if cval == 'median':
            cval = np.nanmedian(image2)
        if cval == 'mean':
            cval = np.nanmean(image2)

        reg_image = self._apply_transform_image(image2, tform, cval=cval)
        logger.info('Filling registered image with cval=%.2f', cval)
        logger.info('Registering image with: '
                    'translation=%s, rotation=%.2f°',
                    tform.translation, np.rad2deg(tform.rotation))
        mask = self._apply_transform_image(mask2, tform, cval=1)
        mask = mask > 0
        return reg_image, mask, tform

    def register_framedata(self, frame1, frame2, cval='median',
                           inplace=False):
        """Align and transform a frame2 to match frame1.

        Parameters
        ----------
        frame1 : `~astropop.framedata.FrameData`
            Reference image for registration.
        image2 : `~astropop.framedata.FrameData`
            Moving image for registration.
        cval : `float` or {'median', 'mean'} (optional)
            Fill value for transformed pixels from outside the image. If
            'median' or 'mean', these statistics will be computed as cval.
            Default: 'median'
        inplace : `bool` (optional)
            Perform the operation in original frame2 instance. The original
            data will be changed.

        Returns
        -------
        reg_frame : `~astropop.framedata.FrameData`
            Registered frame.
        """
        frame1 = check_framedata(frame1)
        frame2 = check_framedata(frame2)

        im1 = np.array(frame1.data)
        im2 = np.array(frame2.data)
        msk1 = frame1.mask if frame1.mask is None else np.array(frame1.mask)
        msk2 = frame2.mask if frame2.mask is None else np.array(frame2.mask)

        data, mask, tform = self.register_image(im1, im2, msk1, msk2,
                                                cval=cval)

        if inplace:
            reg_frame = frame2
        else:
            # Copy the frame to mantain the memmap caching behavior
            reg_frame = frame2.copy()

        reg_frame.data = data
        reg_frame.mask = mask

        if not frame2.uncertainty.empty:
            unct = frame2.get_uncertainty(return_none=False)
            unct = self._apply_transform_image(unct,
                                               tform, cval=np.nan)
            reg_frame.uncertainty = unct

        sx, sy = tform.translation
        theta = np.rad2deg(tform.rotation)
        reg_frame.meta[_keywords['method']] = self._name
        reg_frame.meta[_keywords['shift_x']] = sx
        reg_frame.meta[_keywords['shift_y']] = sy
        reg_frame.meta[_keywords['rotation']] = theta

        if reg_frame.wcs is not None:
            logger.warn('WCS in frame2 is not None. Due to the transform it'
                        ' will be erased.')
            reg_frame.wcs = None

        return reg_frame


class CrossCorrelationRegister(_BaseRegister):
    """Register images usgin `~skimage.Register.phase_cross_correlation`.

    It uses cross-correlation to find a translation-only transform between
    two images. It obtains an initial estimate of the cross-correlation
    peak by an FFT and then refines the shift estimation by upsampling
    the DFT only in a small neighborhood of that estimate by means of a
    matrix-multiply DFT[1]_.

    Parameters
    ----------
    upsample_factor : int (optional)
        Upsampling image factor. Images will be Registered to within
        ``1 / upsample_factor`` of a pixel.
        Default: 1 (no upsampling)
    space : {'real', 'fourier'} (optional)
        Defines how the algorithm interprets input data. "real" means
        data will be FFT'd to compute the correlation, while "fourier"
        data will bypass FFT of input data. Case insensitive.

    References
    ----------
    .. [1] :DOI:`10.1364/OL.33.000156`
    """

    _name = 'cross-correlation'

    def __init__(self, upsample_factor=1, space='real'):
        from skimage.registration import phase_cross_correlation

        self._pcc = partial(phase_cross_correlation, space=space,
                            upsample_factor=upsample_factor,
                            return_error=False)

    def _compute_transform(self, image1, image2, mask1=None, mask2=None):
        if mask1 is not None or mask2 is not None:
            logger.debug("Masks are ignored in CrossCorrelationRegister.")
        # Masks are ignored by default
        dy, dx = self._pcc(image1, image2)
        return transform.AffineTransform(translation=(-dx, -dy))


class AsterismRegister(_BaseRegister):
    """Register images using asterism matching. Based on astroalign [1]_.

    This Register algorith compute the transform between 2 images based
    on the position of detected sources. It can handle both translation and
    rotation of the images. It compare similar 3-points asterisms in 2
    images and find the best possible affine transform between them.

    This package requires `astroalign` to work. The main difference here to
    the bare astroalign Register is that we use our `starfind`
    implementation to find the sources, to keep just good punctual sources
    in the Register, and sort them by brightness, using the brighter sources
    in the work. This may allow a better result in the Register, according our
    experiments.

    Parameters
    ----------
    max_control_points : int, optional
        Maximum control points (stars) used in asterism matching.
        Default: 50
    detection_threshold : int, optional
        Minimum SNR detection threshold.
        Default: 5
    detection_function : {'sepfind', 'starfind', 'daofind'} (optional)
        Detection function to use.
        Default: 'sepfind'
    detection_kwargs : dict (optional)
        Keyword arguments to pass to the detection function.

    Raises
    ------
    ImportError: if astroalign is not installed

    References
    ----------
    .. [1] :DOI:`10.1016/j.ascom.2020.100384`
    """

    _name = 'asterism-matching'

    def __init__(self, max_control_points=50, detection_threshold=5,
                 detection_function='sepfind', **detection_kwargs):
        try:
            import astroalign
        except ImportError:
            raise ImportError('AsterismRegister requires astroalign tools.')

        from ..photometry import sepfind, starfind, daofind, background

        funcs = {'sepfind': sepfind, 'starfind': starfind, 'daofind': daofind}
        self._aa = astroalign
        self._sf = partial(funcs[detection_function], **detection_kwargs)
        self._bkg = background
        self._max_cntl_pts = max_control_points
        self._threshold = detection_threshold

    def _compute_transform(self, image1, image2, mask1=None, mask2=None):
        if mask1 is not None or mask2 is not None:
            logger.debug("Masks are ignored in AsterismRegister.")

        # use our starfind to work with only good sources
        bkg, rms = self._bkg(image1, global_bkg=True)
        sources1 = self._sf(image1, self._threshold, bkg, rms)
        bkg, rms = self._bkg(image2, global_bkg=True)
        sources2 = self._sf(image2, self._threshold, bkg, rms)
        sources1.sort('flux', reverse=True)
        sources2.sort('flux', reverse=True)
        logger.debug('Asterism matching: %d sources in image1, %d in image2',
                     len(sources1), len(sources2))
        sources1 = np.array(list(zip(sources1['x'], sources1['y'])))
        sources2 = np.array(list(zip(sources2['x'], sources2['y'])))

        tform, ctl_pts = self._aa.find_transform(sources1, sources2,
                                                 self._max_cntl_pts)

        logger.debug("Asterism matching performed with sources at: "
                     "image1: %s; image2 %s",
                     ctl_pts[0].tolist(), ctl_pts[1].tolist())

        return tform


def compute_shift_list(frame_list, algorithm='cross-correlation',
                       ref_image=0, skip_failure=False, **kwargs):
    """Compute the shift between a list of frames.

    Parameters
    ----------
    frame_list : list
        A list containing `~astropop.framedata.FrameData` images to be
        registered. All images must have the same shape.
    algorithm : {'cross-correlation', 'asterism-matching'} (optional)
        The algorithm to compute the `~skimage.transform.AffineTransform`
        between the images.
        'cross-correlation' will compute the transform
        using `~skimage.transform.phase_cross_correlation` method.
        'asterism-matching' will use `~astroalign` to match asterisms of 3
        detected stars in the field and compute the transform.
        Default: 'cross-correlation'
    ref_image : int (optional)
        Reference image index to compute the registration.
        Default: 0
    skip_failure : bool (optional)
        If True, the images that fail to register will be skipped and their
        shifts will be set to nan.
        Default: False
    **kwargs :
        keyword arguments to be passed to `CrossCorrelationRegister` or
        `AsterismRegister` during instance creation. See the parameters in
        each class documentation.
    """
    reg = _algorithm_check(algorithm, kwargs)
    _check_compatible_list(frame_list)

    n = len(frame_list)

    ref = frame_list[ref_image]
    ref_im = np.array(ref.data)
    ref_mk = ref.mask if ref.mask is None else np.array(ref.mask)

    shift_list = [None]*n
    for i in range(n):
        logger.info('Computing shift of image %i from %i', i+1, n)
        if i == ref_image:
            shift_list[i] = [0, 0]
            continue

        mov = frame_list[i]
        mov_im = np.array(mov.data)
        mov_mk = mov.mask if mov.mask is None else np.array(mov.mask)
        try:
            tform = reg.compute_transform(ref_im, mov_im, ref_mk, mov_mk)
            shift_list[i] = list(tform.translation)
        except Exception as e:
            logger.warning('Failed to compute shift of image %i: %s', i+1, e)
            if skip_failure:
                shift_list[i] = [np.nan, np.nan]
            else:
                raise e

    return shift_list


def register_framedata_list(frame_list, algorithm='cross-correlation',
                            ref_image=0, clip_output=False,
                            cval='median', inplace=False, skip_failure=False,
                            **kwargs):
    """Perform registration in a framedata list.

    Parameters
    ----------
    frame_list : list
        A list containing `~astropop.framedata.FrameData` images to be
        registered. All images must have the same shape.
    algorith : {'cross-correlation', 'asterism-matching'} (optional)
        The algorithm to compute the `~skimage.transform.AffineTransform`
        between the images.
        'cross-correlation' will compute the transform
        using `~skimage.transform.phase_cross_correlation` method.
        'asterism-matching' will use `~astroalign` to match asterisms of 3
        detected stars in the field and compute the transform.
        Default: 'cross-correlation'
    ref_image : int (optional)
        Reference image index to compute the registration.
        Default: 0
    clip_output : bool (optional)
        If True, the output images will be clipped to a only-valid pixels
        frame.
    cval : float or {'median', 'mean'} (optional)
        Fill value for the empty pixels in the transformed image. If 'mean' or
        'median', the correspondent values will be computed from the image.
        Default: 'median'
    skip_failure: bool (optional)
        If True, the images that fail to register will be skipped. Their data
        will be fill with the cval and all pixels mask will be set to
        True. If False, the error will be raised.
        Default: False
    inplace : bool (optional)
        Perform the operation inplace, modifying the original FrameData
        container. If `False`, a new container will be created.
        Default: `False`
    **kwargs :
        keyword arguments to be passed to `CrossCorrelationRegister` or
        `AsterismRegister` during instance creation. See the parameters in
        each class documentation.
    """
    reg = _algorithm_check(algorithm, kwargs)
    _check_compatible_list(frame_list)

    n = len(frame_list)
    reg_list = [None]*n
    for i in range(n):
        logger.info('Registering image %i from %i', i+1, n)
        try:
            reg_list[i] = reg.register_framedata(frame_list[ref_image],
                                                 frame_list[i],
                                                 cval=cval, inplace=inplace)
        except Exception as e:
            if not skip_failure:
                raise e
            logger.warning('Failed to register image %i: %s', i+1, e)
            reg_list[i] = check_framedata(frame_list[i], copy=not inplace)
            if cval == 'median':
                icval = np.median(reg_list[i].data)
            elif cval == 'mean':
                icval = np.mean(reg_list[i].data)
            else:
                icval = cval
            reg_list[i].data[:] = icval
            reg_list[i].mask[:] = True
            reg_list[i].meta[_keywords['method']] = 'failed'
            reg_list[i].meta[_keywords['shift_x']] = None
            reg_list[i].meta[_keywords['shift_y']] = None
            reg_list[i].meta[_keywords['rotation']] = None

    if clip_output:
        shifts = [(i.meta[_keywords['shift_x']], i.meta[_keywords['shift_y']])
                  if i.meta[_keywords['method']] != 'failed'
                  else (np.nan, np.nan)
                  for i in reg_list]
        shifts = np.array(shifts)
        xslice, yslice = _get_clip_slices(shifts, reg_list[0].shape)
        logger.info('Clipping output images to section: x=%s:%s, y=%s:%s',
                    xslice.start, xslice.stop, yslice.start, yslice.stop)
        for i in range(n):
            trim_image(reg_list[i], xslice, yslice, inplace=True)

    return reg_list



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Catalog managing and query."""

from .vizier import VizierCatalogClass, \
                    UCAC4Catalog, UCAC5Catalog, GSC23Catalog, \
                    APASSCalatolg, DENISCatalog, TWOMASSCatalog
from .simbad import SimbadCatalog, SimbadCatalogClass
from .local import TableCatalog, ASCIICatalog, FITSCatalog
from .utils import identify_stars


__all__ = ['identify_stars', 'catalogs_available',
           'VizierCatalogClass', 'SimbadCatalogClass', 'UCAC5Catalog',
           'SimbadCatalog', 'UCAC4Catalog', 'GSC23Catalog', 'APASSCalatolg',
           'DENISCatalog', 'TWOMASSCatalog', 'default_catalogs',
           'TableCatalog', 'ASCIICatalog', 'FITSCatalog']


default_catalogs = {'Simbad': SimbadCatalog,
                    'UCAC4': UCAC4Catalog,
                    'UCAC5': UCAC5Catalog,
                    'APASS': APASSCalatolg,
                    'DENIS': DENISCatalog,
                    '2MASS': TWOMASSCatalog,
                    'GSC2.3': GSC23Catalog}


catalogs_available = default_catalogs.keys()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
import numpy as np
from astropop.framedata.compat import extract_header_wcs, _extract_ccddata, \
                                      _extract_fits, _merge_and_clean_header
from astropy.io import fits
from astropy.wcs import WCS
from astropy.table import Table
from astropy.nddata import StdDevUncertainty, InverseVariance, \
                           VarianceUncertainty, UnknownUncertainty, \
                           CCDData

from astropop.logger import logger, log_to_list
from astropop.testing import *


_comon_wcs_keys = ('CTYPE', 'CRVAL', 'CRPIX', 'CD1_', 'CD2_', 'PC1_', 'PC2_')


_base_header = """SIMPLE  =                    T / Fits standard
BITPIX  =                  -32 / FOUR-BYTE SINGLE PRECISION FLOATING POINT
NAXIS   =                    2 / STANDARD FITS FORMAT
NAXIS1  =                  256 / STANDARD FITS FORMAT
NAXIS2  =                  256 / STANDARD FITS FORMAT
DATE-OBS= '1996-10-14T10:14:36.123' / Date and time of start of obs. in UTC.
"""

_wcs_no_sip = """
CTYPE1  = 'RA---TAN'       / Tangent projection
CTYPE2  = 'DEC--TAN'       / Tangent projection
CRVAL1  =     202.482322805429 / [deg] RA at CRPIX1,CRPIX2
CRVAL2  =     47.1751189300101 / [deg] DEC at CRPIX1,CRPIX2
CRPIX1  =                 128. / Reference pixel along axis 1
CRPIX2  =                 128. / Reference pixel along axis 2
CD1_1   = 0.000249756880272355 / Corrected CD matrix
CD1_2   = 0.000230177809743655 / Corrected CD matrix
CD2_1   = 0.000230428519265417 / Corrected CD matrix element
CD2_2   = -0.000249965770576587 / Corrected CD matrix element
"""

_wcs_sip = """
CTYPE1  = 'RA---TAN-SIP'       / RA---TAN with distortion in pixel
CTYPE2  = 'DEC--TAN-SIP'       / DEC--TAN with distortion in pixel
CRVAL1  =     202.482322805429 / [deg] RA at CRPIX1,CRPIX2
CRVAL2  =     47.1751189300101 / [deg] DEC at CRPIX1,CRPIX2
CRPIX1  =                 128. / Reference pixel along axis 1
CRPIX2  =                 128. / Reference pixel along axis 2
CD1_1   = 0.000249756880272355 / Corrected CD matrix element
CD1_2   = 0.000230177809743655 / Corrected CD matrix element
CD2_1   = 0.000230428519265417 / Corrected CD matrix element
CD2_2   = -0.000249965770576587 / Corrected CD matrix element
A_ORDER =                    3 / polynomial order, axis 1, detector to sky
A_0_2   =           2.9656E-06 / distortion coefficient
A_0_3   =           3.7746E-09 / distortion coefficient
A_1_1   =           2.1886E-05 / distortion coefficient
A_1_2   =          -1.6847E-07 / distortion coefficient
A_2_0   =          -2.3863E-05 / distortion coefficient
A_2_1   =           -8.561E-09 / distortion coefficient
A_3_0   =          -1.4172E-07 / distortion coefficient
A_DMAX  =                1.394 / [pixel] maximum correction
B_ORDER =                    3 / polynomial order, axis 2, detector to sky
B_0_2   =             2.31E-05 / distortion coefficient
B_0_3   =          -1.6168E-07 / distortion coefficient
B_1_1   =          -2.4386E-05 / distortion coefficient
B_1_2   =          -5.7813E-09 / distortion coefficient
B_2_0   =           2.1197E-06 / distortion coefficient
B_2_1   =          -1.6583E-07 / distortion coefficient
B_3_0   =          -2.0249E-08 / distortion coefficient
B_DMAX  =                1.501 / [pixel] maximum correction
AP_ORDER=                    3 / polynomial order, axis 1, sky to detector
AP_0_1  =          -6.4275E-07 / distortion coefficient
AP_0_2  =          -2.9425E-06 / distortion coefficient
AP_0_3  =           -3.582E-09 / distortion coefficient
AP_1_0  =          -1.4897E-05 / distortion coefficient
AP_1_1  =           -2.225E-05 / distortion coefficient
AP_1_2  =           1.7195E-07 / distortion coefficient
AP_2_0  =           2.4146E-05 / distortion coefficient
AP_2_1  =            6.709E-09 / distortion coefficient
AP_3_0  =           1.4492E-07 / distortion coefficient
BP_ORDER=                    3 / polynomial order, axis 2, sky to detector
BP_0_1  =          -1.6588E-05 / distortion coefficient
BP_0_2  =          -2.3424E-05 / distortion coefficient
BP_0_3  =            1.651E-07 / distortion coefficient
BP_1_0  =          -2.6783E-06 / distortion coefficient
BP_1_1  =           2.4753E-05 / distortion coefficient
BP_1_2  =           3.8917E-09 / distortion coefficient
BP_2_0  =           -2.151E-06 / distortion coefficient
BP_2_1  =              1.7E-07 / distortion coefficient
BP_3_0  =           2.0482E-08 / distortion coefficient
"""

_invalid_wcs = """
CRVAL1  =         164.98110962 / Physical value of the reference pixel X
CRVAL2  =          44.34089279 / Physical value of the reference pixel Y
CRPIX1  =                -34.0 / Reference pixel in X (pixel)
CRPIX2  =               2041.0 / Reference pixel in Y (pixel)
CDELT1  =           0.10380000 / X Scale projected on detector (#/pix)
CDELT2  =           0.10380000 / Y Scale projected on detector (#/pix)
CTYPE1  = 'RA---TAN'           / Pixel coordinate system
CTYPE2  = 'WAVELENGTH'         / Pixel coordinate system
CUNIT1  = 'degree  '           / Units used in both CRVAL1 and CDELT1
CUNIT2  = 'nm      '           / Units used in both CRVAL2 and CDELT2
CD1_1   =           0.20760000 / Pixel Coordinate translation matrix
CD1_2   =           0.00000000 / Pixel Coordinate translation matrix
CD2_1   =           0.00000000 / Pixel Coordinate translation matrix
CD2_2   =           0.10380000 / Pixel Coordinate translation matrix
C2YPE1  = 'RA---TAN'           / Pixel coordinate system
C2YPE2  = 'DEC--TAN'           / Pixel coordinate system
C2NIT1  = 'degree  '           / Units used in both C2VAL1 and C2ELT1
C2NIT2  = 'degree  '           / Units used in both C2VAL2 and C2ELT2
RADECSYS= 'FK5     '           / The equatorial coordinate system
"""

_hist_comm_blank = """
HISTORY First line of history
HISTORY Second line of history
HISTORY Third line of history
COMMENT This is a first comment
COMMENT This is a second comment
COMMENT This is a third comment
"""

class Test_ExtractHeader():
    def test_merge_and_clean_header(self):
        strhdr = _base_header+_wcs_no_sip+_hist_comm_blank
        header = fits.Header.fromstring(strhdr, sep='\n')
        header['TEST'] = ('value', 'comment')

        meta = {'META1': 1, 'META2': 2}
        meta, wcs, history, comments = _merge_and_clean_header(meta, header,
                                                               None)
        assert_is_instance(meta, fits.Header)
        assert_equal(meta['META1'], 1)
        assert_equal(meta['META2'], 2)
        assert_equal(meta['TEST'], 'value')
        for i in ['history', 'comment', '']:
            assert_not_in(i, meta)

        assert_is_instance(wcs, WCS)
        assert_equal(wcs.wcs.ctype, ['RA---TAN', 'DEC--TAN'])
        assert_equal(wcs.wcs.cunit, ['degree', 'degree'])
        assert_equal(wcs.wcs.crval, (202.482322805429, 47.1751189300101))
        assert_equal(wcs.wcs.crpix, (128, 128))
        assert_equal(wcs.wcs.cdelt, (1, 1))
        assert_equal(wcs.wcs.dateobs, '1996-10-14T10:14:36.123')

        assert_is_instance(history, list)
        assert_is_instance(comments, list)
        assert_equal(history, ['First line of history',
                               'Second line of history',
                               'Third line of history'])
        assert_equal(comments, ['This is a first comment',
                                'This is a second comment',
                                'This is a third comment'])

    def test_extract_header_nowcs(self):
        header = fits.Header.fromstring(_base_header, sep='\n')
        h, wcs = extract_header_wcs(header)
        assert_is_none(wcs)
        assert_is_instance(h, fits.Header)
        assert_equal(h, header)
        assert_false(h is header)

    def test_extract_header_nosip(self):
        header = fits.Header.fromstring(_base_header+_wcs_no_sip, sep='\n')
        h, wcs = extract_header_wcs(header)
        assert_is_instance(wcs, WCS)
        assert_equal(wcs.wcs.ctype[0], 'RA---TAN')
        assert_equal(wcs.wcs.ctype[1], 'DEC--TAN')
        assert_is_instance(h, fits.Header)
        for i in _comon_wcs_keys:
            assert_not_in(f'{i}1', h.keys())
            assert_not_in(f'{i}2', h.keys())
        assert_in('DATE-OBS', h.keys())
        assert_false(h is header)
        assert_not_equal(h, header)

    def test_extract_header_sip(self):
        header = fits.Header.fromstring(_base_header+_wcs_sip, sep='\n')
        h, wcs = extract_header_wcs(header)
        assert_is_instance(wcs, WCS)
        assert_equal(wcs.wcs.ctype[0], 'RA---TAN-SIP')
        assert_equal(wcs.wcs.ctype[1], 'DEC--TAN-SIP')
        assert_is_instance(h, fits.Header)
        for i in _comon_wcs_keys:
            assert_not_in(f'{i}1', h.keys())
            assert_not_in(f'{i}2', h.keys())
        for i in ('A_0_2', 'AP_2_0', 'BP_ORDER', 'A_DMAX'):
            assert_not_in(i, h.keys())
        assert_in('DATE-OBS', h.keys())
        assert_false(h is header)
        assert_not_equal(h, header)

    def test_extract_invalid_wcs_header(self):
        # It should no raise, just return empty wcs
        # No header change too
        header = fits.Header.fromstring(_base_header+_invalid_wcs, sep='\n')
        del header['']
        h, wcs = extract_header_wcs(header)

        assert_is_none(wcs)
        assert_is_instance(h, fits.Header)
        assert_equal(h, header)
        assert_false(h is header)


class Test_Extract_CCDData():
    shape = (10, 10)
    meta = {'observer': 'testing', 'very long keyword': 1}
    unit = 'adu'

    @property
    def mask(self):
        mask = np.zeros(self.shape, dtype=bool)
        mask[1:3, 1:3] = 1
        return mask

    @property
    def ccd(self):
        data = 100*np.ones(self.shape)
        uncert = StdDevUncertainty(np.sqrt(data), unit=self.unit)
        unit = self.unit
        return CCDData(data, uncertainty=uncert, mask=self.mask,
                       unit=unit, meta=self.meta)

    def test_simple_import(self):
        ccd = self.ccd
        f = _extract_ccddata(ccd)
        assert_equal(f['data'], 100*np.ones(self.shape))
        assert_equal(f['mask'], self.mask)
        assert_equal(f['unit'], self.unit)
        assert_equal(f['meta'], self.meta)
        assert_equal(f['uncertainty'], 10*np.ones(self.shape))

    def test_variance_uncert(self):
        ccd = self.ccd
        uncert = VarianceUncertainty(100*np.ones(self.shape),
                                     unit=self.unit+'2')
        ccd.uncertainty = uncert

        f = _extract_ccddata(ccd)
        assert_equal(f['data'], 100*np.ones(self.shape))
        assert_equal(f['mask'], self.mask)
        assert_equal(f['unit'], self.unit)
        assert_equal(f['meta'], self.meta)
        assert_equal(f['uncertainty'], 10*np.ones(self.shape))

    def test_inverse_variance_uncert(self):
        ccd = self.ccd
        uncert = InverseVariance(0.01*np.ones(self.shape),
                                 unit=self.unit+'-2')
        ccd.uncertainty = uncert

        f = _extract_ccddata(ccd)
        assert_equal(f['data'], 100*np.ones(self.shape))
        assert_equal(f['mask'], self.mask)
        assert_equal(f['unit'], self.unit)
        assert_equal(f['meta'], self.meta)
        assert_equal(f['uncertainty'], 10*np.ones(self.shape))

    def test_no_mask(self):
        ccd = self.ccd
        ccd.mask = None
        f = _extract_ccddata(ccd)
        assert_is_none(f['mask'])

    def test_no_uncertainty(self):
        ccd = self.ccd
        ccd.uncertainty = None
        f = _extract_ccddata(ccd)
        assert_is_none(f['uncertainty'])

    def test_invalid_uncert(self):
        ccd = self.ccd
        ccd.uncertainty = UnknownUncertainty(np.ones(self.shape))

        with pytest.raises(TypeError):
            _extract_ccddata(ccd)

    def test_invalid_uncert_unit(self):
        ccd = self.ccd
        # we must do this to avoid ccddata self protection
        uncert = StdDevUncertainty(np.sqrt(ccd.data), unit='m')
        ccd._uncertainty = uncert
        ccd._uncertainty._parent_nddata = ccd

        with pytest.raises(ValueError):
            _extract_ccddata(ccd)


class Test_Fits_Extract():
    shape = (10, 10)

    @property
    def mask(self):
        mask = np.zeros(self.shape, dtype=bool)
        mask[1:3, 1:3] = 1
        return mask

    def create_hdu(self, uncert=False, mask=False,
                   unit='adu', unit_key='BUNIT'):
        l = []
        data = 100*np.ones(self.shape)
        header = self.create_header(unit=unit, unit_key=unit_key)
        data_hdu = fits.PrimaryHDU(data, header=header)
        l.append(data_hdu)

        if uncert:
            uncert_hdu = fits.ImageHDU(np.ones(self.shape), name='UNCERT')
            l.append(uncert_hdu)

        if mask:
            mask_hdu = fits.ImageHDU(self.mask.astype('uint8'), name='MASK')
            l.append(mask_hdu)

        hdul = fits.HDUList(l)
        return hdul

    def create_header(self, unit=None, unit_key=None):
        header = fits.Header()
        if unit is not None:
            header[unit_key] = unit
        return header

    def test_simple_hdu(self):
        hdu = self.create_hdu()[0]
        header = self.create_header('adu', 'BUNIT')
        f = _extract_fits(hdu)
        assert_equal(f['data'], 100*np.ones(self.shape))
        for i in header:
            assert_equal(f['meta'][i], header[i])
        assert_equal(f['unit'], 'adu')

    def test_simple_hdulist(self):
        hdu = self.create_hdu()
        header = self.create_header('adu', 'BUNIT')
        f = _extract_fits(hdu)
        assert_equal(f['data'], 100*np.ones(self.shape))
        for i in header:
            assert_equal(f['meta'][i], header[i])
        assert_equal(f['unit'], 'adu')

    def test_hdulist_with_uncertainty(self):
        hdu = self.create_hdu(uncert=True)
        header = self.create_header('adu', 'BUNIT')
        f = _extract_fits(hdu, hdu_uncertainty='UNCERT')
        assert_equal(f['data'], 100*np.ones(self.shape))
        for i in header:
            assert_equal(f['meta'][i], header[i])
        assert_equal(f['unit'], 'adu')
        assert_equal(f['uncertainty'], np.ones(self.shape))

    def test_hdulist_with_mask(self):
        hdu = self.create_hdu(mask=True)
        header = self.create_header('adu', 'BUNIT')
        f = _extract_fits(hdu, hdu_mask='MASK')
        assert_equal(f['data'], 100*np.ones(self.shape))
        for i in header:
            assert_equal(f['meta'][i], header[i])
        assert_equal(f['unit'], 'adu')
        assert_equal(f['mask'], self.mask.astype('uint8'))

    def test_hdulist_full(self):
        hdu = self.create_hdu(uncert=True, mask=True)
        header = self.create_header('adu', 'BUNIT')
        f = _extract_fits(hdu, hdu_mask='MASK')
        assert_equal(f['data'], 100*np.ones(self.shape))
        for i in header:
            assert_equal(f['meta'][i], header[i])
        assert_equal(f['unit'], 'adu')
        assert_equal(f['uncertainty'], np.ones(self.shape))
        assert_equal(f['mask'], self.mask.astype('uint8'))

    def test_invalid_unit(self):
        hdu = self.create_hdu(unit='invalid')
        with pytest.raises(ValueError):
            _extract_fits(hdu)

    def test_file(self, tmpdir):
        hdu = self.create_hdu(uncert=True, mask=True)
        header = self.create_header('adu', 'BUNIT')
        fname = tmpdir.join('test_file.fits').strpath
        hdu.writeto(fname)

        f = _extract_fits(fname)

        assert_equal(f['data'], 100*np.ones(self.shape))
        for i in header:
            assert_equal(f['meta'][i], header[i])
        assert_equal(f['unit'], 'adu')
        assert_equal(f['uncertainty'], np.ones(self.shape))
        assert_equal(f['mask'], self.mask.astype('uint8'))

    def test_data_in_other_hdu(self, tmpdir):
        tbl = Table(np.ones(10).reshape(5, 2))
        data = 100*np.ones(self.shape)
        hdul = fits.HDUList(hdus=[fits.PrimaryHDU(),
                                  fits.TableHDU(tbl.as_array()),
                                  fits.ImageHDU(data)])
        fname = tmpdir.join('test_table.fits').strpath
        hdul.writeto(fname)

        logs = []
        lh = log_to_list(logger, logs, full_record=True)
        f = _extract_fits(fname)
        assert_equal(f['data'], 100*np.ones(self.shape))
        assert_equal(f['unit'], None)

        # ensure log emitting
        logs = [i for i in logs if i.message == 'First hdu with image data: 2']
        assert_equal(len(logs), 1)
        assert_equal(logs[0].levelname, 'INFO')

        logger.removeHandler(lh)

    def test_invalid_type(self):
        with pytest.raises(TypeError):
            _extract_fits(None)

        with pytest.raises(TypeError):
            _extract_fits(np.array([1, 2, 3]))

        with pytest.raises(TypeError):
            _extract_fits(CCDData([1, 2, 3], unit='adu'))

    def test_no_data_in_hdu(self):
        hdul = fits.HDUList([fits.PrimaryHDU(), fits.ImageHDU()])
        with pytest.raises(ValueError):
            _extract_fits(hdul, hdu=1)

    def test_unit_incompatible(self):
        hdul = self.create_hdu(uncert=True, mask=True)
        hdul[0].header['BUNIT'] = 'm'

        with pytest.raises(ValueError):
            _extract_fits(hdul, hdu_uncertainty='UNCERT',
                          unit_key='BUNIT', unit='adu')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Process astrometry calibration using several algorithms."""

from .astrometrynet import AstrometryNetUnsolvedField, \
                           AstrometrySolver, \
                           solve_astrometry_image, \
                           solve_astrometry_xy, \
                           create_xyls
from .manual_wcs import wcs_from_coords
from .coords_utils import guess_coordinates

__all__ = ['guess_coordinates', 'wcs_from_coords', 'AstrometrySolver',
           'solve_astrometry_xy', 'solve_astrometry_image',
           'create_xyls', 'AstrometryNetUnsolvedField']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Automated alibration of photometry with reference catalogs."""

import numpy as np

__all__ = ['solve_photometry_montecarlo', 'solve_photometry_median',
           'solve_photometry_average']


def _scale_operator(measure_scale, out_scale):
    """Fix units problemas and get correct scale."""
    diff = None  # find the out/measure ratio (- for mag)
    corr = None  # correct the measure by diff (+ for mag)
    trans_flux = None  # transform the measure to out scale
    # error_func = None  # error of the final calculation
    if out_scale == 'mag':
        diff = np.subtract
        corr = np.add
        if measure_scale == 'linear':
            def trans_flux(x): return -2.5*np.log10(x) + 25
            def error_func(x, x_e): return 1.086*((x_e + np.sqrt(x))/x)
        elif measure_scale == 'log':
            def trans_flux(x): return -2.5*x + 25
            error_func = None
        elif measure_scale == 'mag':
            def trans_flux(x): return x
            error_func = None
    elif out_scale == 'log':
        diff = np.subtract
        corr = np.add
        if measure_scale == 'linear':
            def trans_flux(x): return np.log10(x)
            error_func = None
        elif measure_scale == 'log':
            def trans_flux(x): return x
            error_func = None
        elif measure_scale == 'mag':
            def trans_flux(x): return x/(-2.5) + 10
            error_func = None
    elif out_scale == 'linear':
        diff = np.divide
        corr = np.multiply
        if measure_scale == 'linear':
            def trans_flux(x): return x
            error_func = None
        elif measure_scale == 'log':
            def trans_flux(x): return 10**x
            error_func = None
        elif measure_scale == 'mag':
            def trans_flux(x): return 10**(-0.4*x + 10)
            error_func = None

    return trans_flux, diff, corr, error_func


def solve_photometry_median(fluxes, flux_error, references, limits=(5, 18),
                            flux_scale='linear', ref_scale='mag'):
    """Solve the photometry by the median comparison of field stars."""
    trans_func, diff_func, corr_func, error_func = _scale_operator(flux_scale,
                                                                   ref_scale)
    mags = trans_func(fluxes)

    a, b = limits
    a, b = a, b if a < b else b, a
    args = np.where(np.logical_and(references >= a, references <= b))

    diff = diff_func(references, mags)
    dif = np.nanmedian(diff[args])
    err = np.nanstd(diff[args])

    # now we assume just the statistical error
    error = err
    return corr_func(mags, dif), error


def solve_photometry_average(fluxes, flux_error, references, limits=(5, 18),
                             flux_scale='linear', ref_scale='mag'):
    """Solve the photometry by the average comparison of field stars."""
    trans_func, diff_func, corr_func, error_func = _scale_operator(flux_scale,
                                                                   ref_scale)
    mags = trans_func(fluxes)

    a, b = limits
    a, b = a, b if a < b else b, a
    args = np.where(np.logical_and(references >= a, references <= b))

    diff = diff_func(references, mags)
    dif = np.nanaverage(diff[args], weights=np.divide(1, flux_error[args]))
    err = np.nanstd(diff[args])

    error = err
    return corr_func(mags, dif), error


def _montecarlo_loop(args):
    mags = args[0]
    references = args[1]
    n_stars = args[2]
    diff_func = args[3]
    corr_func = args[4]

    iter_mags = np.zeros(len(mags))
    iter_mags[:] = np.nan

    choices = np.random.choice(len(mags), n_stars)
    iter_mags = np.nanmedian(diff_func(references[choices], mags[choices]))
    return corr_func(mags, iter_mags)


def solve_photometry_montecarlo(fluxes, flux_error, ref_mags, limits=(5, 18),
                                n_iter=100, n_stars=0.5,
                                flux_scale='linear', ref_scale='mag'):
    """Solve the photometry using montecarlo comparison of field stars."""
    trans_func, diff_func, corr_func, error_func = _scale_operator(flux_scale,
                                                                   ref_scale)
    mags = trans_func(fluxes)

    if float(n_stars).is_integer():
        n_stars = n_stars
    else:
        n_stars = max(1, int(n_stars*len(fluxes)))

    nrefs = np.array(ref_mags)
    lim = sorted(limits)
    filt = np.where(np.logical_or(nrefs < lim[0], nrefs > lim[1]))
    nrefs[filt] = np.nan

    args = (mags, nrefs, n_stars, diff_func, corr_func)
    iter_mags = [_montecarlo_loop(args) for i in range(n_iter)]

    result = np.nanmedian(iter_mags, axis=0)
    errors = np.nanstd(iter_mags, axis=0)
    if error_func is not None:
        errors = np.sqrt(errors**2 + error_func(fluxes, flux_error)**2)
    return result, errors



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Gaussian models and utilities."""
import numpy as np
from astropy.modeling import Fittable1DModel, Fittable2DModel, Parameter
from astropy.stats import gaussian_sigma_to_fwhm


__all__ = ['gaussian_r', 'gaussian_1d', 'gaussian_2d', 'GaussianRadial',
           'PSFGaussian1D', 'PSFGaussian2D']


def gaussian_normalize(sigma, sigma2=None):
    """Normalize the gaussian distribution."""
    if sigma2 is None:
        sigma2 = sigma
    return 1.0/(np.sqrt(2*np.pi*sigma*sigma2))


def gaussian_fwhm(sigma, sigma2=None):
    """Get FWHM based on gaussian sigma."""
    if sigma2 is not None:
        return np.abs(gaussian_sigma_to_fwhm*np.hypot(sigma, sigma2))
    else:
        return np.abs(gaussian_sigma_to_fwhm*sigma)


def gaussian_r(r, sigma, flux, sky):
    a = flux*gaussian_normalize(sigma)
    return sky + a*np.exp(-0.5*(r/sigma)**2)


def gaussian_1d(x, x0, sigma, flux, sky):
    a = flux*gaussian_normalize(sigma)
    return sky + a*np.exp(-0.5*((x-x0)/sigma)**2)


def gaussian_2d(x, y, x0, y0, sigma_x, sigma_y, theta, flux, sky):
    cost2 = np.cos(np.radians(theta))**2
    sint2 = np.sin(np.radians(theta))**2
    sin2t = np.sin(2*np.radians(theta))
    xstd2 = sigma_x**2
    ystd2 = sigma_y**2
    a = (cost2/xstd2) + (sint2/ystd2)
    b = (sin2t/xstd2) - (sin2t/ystd2)
    c = (sint2/xstd2) + (cost2/ystd2)
    xi = x - x0
    yi = y - y0
    amp = flux*gaussian_normalize(sigma_x, sigma_y)
    return amp*np.exp(-0.5*((a*xi**2) + (b*xi*yi) + (c*yi**2)))


class PSFGaussian2D(Fittable2DModel):
    flux = Parameter(default=1, fixed=False)
    x_0 = Parameter(default=0)
    y_0 = Parameter(default=0)
    sigma_x = Parameter(default=1, fixed=False)
    sigma_y = Parameter(default=1, fixed=False)
    theta = Parameter(default=1, fixed=False)
    sky = Parameter(default=0, fixed=False)

    @staticmethod
    def evaluate(x, y, x_0, y_0, sigma_x, sigma_y, theta, flux, sky):
        return gaussian_2d(x, y, x_0, y_0, sigma_x, sigma_y, theta, flux, sky)

    @property
    def fwhm(self):
        return gaussian_fwhm(self.sigma_x, self.sigma_y)


class PSFGaussian1D(Fittable1DModel):
    flux = Parameter(default=1, fixed=False)
    x_0 = Parameter(default=0)
    sigma = Parameter(default=1, fixed=False)
    sky = Parameter(default=0, fixed=False)

    @staticmethod
    def evaluate(x, x_0, sigma, flux, sky):
        return gaussian_1d(x, x_0, sigma, flux, sky)

    @property
    def fwhm(self):
        return gaussian_fwhm(self.sigma)


class GaussianRadial(Fittable1DModel):
    flux = Parameter(default=1, fixed=False)
    sigma = Parameter(default=1, fixed=False)
    sky = Parameter(default=0, fixed=False)

    @staticmethod
    def evaluate(x, sigma, flux, sky):
        return gaussian_r(x, sigma, flux, sky)

    @property
    def fwhm(self):
        return gaussian_fwhm(self.sigma)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Small utility to identify stars in a catalog."""

from astropy.table import Table

from ..py_utils import process_list, string_fix

__all__ = ['identify_stars']


def identify_stars(x, y, wcs, catalog, **kwargs):
    """Identify and name stars based on their x,y positions and WCS.

    Parameters
    ----------
    x, y: array_like
        x and y positions ofthe stars in the frame.
    wcs: `~astropy.wcs.WCS`
        World Coordinate System reference to identify the astronomical
        coordinates ofthe stars.
    catalog: `~astropop.catalogs._BaseCatalog`
        Catalog for star name matching.
    science_catalog: `~astropop.catalogs._BaseCatalog`
        Custom catalog for star name matching.
    band: `str` (optional)
        Photometric filter for flux estimation.
    limit_angle: `float`, `str` or `~astropy.coordinates.Angle` (optional)
        Maximum search limit to identify the star.
        Default: '2 arcsec'

    Returns
    -------
    `~astropy.table.Table`: All stars with id matched in the catalog.
    """
    limit_angle = kwargs.get('limit_angle', '2 arcsec')
    band = kwargs.get('band')
    cat = catalog
    ra, dec = wcs.all_pix2world(x, y, 1)  # Native wcs has 1 index counting

    cat_res = cat.match_objects(ra, dec, band, limit_angle=limit_angle)

    res = Table()
    res['ra'] = ra
    res['dec'] = dec
    res['cat_id'] = process_list(string_fix, cat_res['id'])
    res['cat_mag'] = cat_res['flux']
    res['cat_mag_err'] = cat_res['flux_error']

    if 'science_catalog' in kwargs.keys():
        sci = kwargs.get('science_catalog')
        sci_res = sci.match_objects(ra, dec, limit_angle=limit_angle)
        res['sci_id'] = process_list(string_fix, sci_res['id'])

    return res



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Manage SQL databases in a simplier way."""

import sqlite3 as sql
import numpy as np
from astropy.table import Table

from .logger import logger
from .py_utils import check_iterable, broadcast


__all__ = ['SQLDatabase', 'SQLTable', 'SQLRow', 'SQLColumn', 'SQLColumnMap']


_ID_KEY = '__id__'


class SQLColumnMap:
    """Map keywords to SQL columns."""

    def __init__(self, db, map_table, map_key, map_column):
        self.db = db
        self.map = db[map_table]
        self.key = map_key
        self.col = map_column

        self._clear_cache()

    def add_column(self, name):
        """Add a new column to the table."""
        name = name.lower()

        if name in self.keywords:
            raise ValueError(f'{name} already exists')

        i = len(self.keywords)+1
        col = f'col_{i}'
        while col in self.keywords:
            i += 1
            col = f'col_{i}'

        self.map.add_rows({self.key: name, self.col: col})
        self._clear_cache()
        return col

    def get_column_name(self, item, add_columns=False):
        """Get the column name for a given keyword."""
        if check_iterable(item):
            return [self.get_column_name(i) for i in item]

        item = item.lower()
        if item not in self.keywords:
            if add_columns:
                return self.add_column(item)
            raise KeyError(f'{item}')

        return self.columns[np.where(self.keywords == item)][0]

    def get_keyword(self, item):
        """Get the keyword for a given column."""
        if check_iterable(item):
            return [self.get_keyword(i) for i in item]

        item = item.lower()
        if item not in self.columns:
            raise KeyError(f'{item}')

        return self.keywords[np.where(self.columns == item)][0]

    def _clear_cache(self):
        self._columns = None
        self._keywords = None

    @property
    def columns(self):
        """Get the column names for the table."""
        if self._columns is None:
            self._columns = np.array(self.map.select(columns=[self.col]))
        return self._columns

    @property
    def keywords(self):
        """Get the keywords of the columns for the table."""
        if self._keywords is None:
            self._keywords = np.array(self.map.select(columns=[self.key]))
        return self._keywords

    def map_row(self, data, add_columns=False):
        """Map a row to the columns."""
        if isinstance(data, dict):
            d = {}
            for k, v in data.items():
                if k in self.keywords or add_columns:
                    d[self.get_column_name(k, add_columns=add_columns)] = v
            data = d
        elif not isinstance(data, list):
            raise ValueError('Only dict and list are supported')
        return data

    def parse_where(self, where):
        """Parse a where clause using column mappring."""
        if isinstance(where, dict):
            return {self.get_column_name(k): v for k, v in where.items()}
        raise TypeError('Only dict is supported')


class SQLTable:
    """Handle an SQL table operations interfacing with the DB."""

    def __init__(self, db, name, colmap=None):
        """Initialize the table.

        Parameters
        ----------
        db : SQLDatabase
            The parent database object.
        name : str
            The name of the table in the database.
        """
        self._db = db
        self._name = name
        self._colmap = colmap

    @property
    def name(self):
        """Get the name of the table."""
        return self._name

    @property
    def db(self):
        """Get the database name."""
        return self._db._db

    @property
    def column_names(self):
        """Get the column names of the current table."""
        names = self._db.column_names(self._name)
        if self._colmap is not None:
            return self._colmap.get_keyword(names)
        return names

    @property
    def values(self):
        """Get the values of the current table."""
        return self.select()

    def select(self, **kwargs):
        """Select rows from the table."""
        where = kwargs.pop('where', None)
        order = kwargs.pop('order', None)
        if self._colmap is not None:
            if where is not None:
                where = self._colmap.parse_where(where)
            if order is not None:
                order = self._colmap.get_column_name(order)

        return self._db.select(self._name, where=where, order=order, **kwargs)

    def as_table(self):
        """Return the current table as an `~astropy.table.Table` object."""
        if len(self) == 0:
            return Table(names=self.column_names)
        return Table(rows=self.values,
                     names=self.column_names)

    def add_column(self, name, data=None):
        """Add a column to the table."""
        if self._colmap is not None:
            name = self._colmap.add_column(name)
        self._db.add_column(self._name, name, data=data)

    def add_rows(self, data, add_columns=False):
        """Add a row to the table."""
        # If keymappging is used, only dict and list
        if self._colmap is not None:
            data = self._colmap.map_row(data, add_columns=add_columns)
        self._db.add_rows(self._name, data, add_columns=add_columns)

    def get_column(self, column):
        """Get a given column from the table."""
        if self._colmap is not None:
            column = self._colmap.get_column_name(column)
        return self._db.get_column(self._name, column)

    def get_row(self, row):
        """Get a given row from the table."""
        return self._db.get_row(self._name, row, column_map=self._colmap)

    def set_column(self, column, data):
        """Set a given column in the table."""
        if self._colmap is not None:
            column = self._colmap.get_column_name(column)
        self._db.set_column(self._name, column, data)

    def set_row(self, row, data):
        """Set a given row in the table."""
        if self._colmap is not None:
            data = self._colmap.map_row(data)
        self._db.set_row(self._name, row, data)

    def index_of(self, where):
        """Get the index of the rows that match the given condition."""
        if self._colmap is not None:
            where = self._colmap.parse_where(where)
        return self._db.index_of(self._name, where)

    def _resolve_tuple(self, key):
        """Resolve how tuples keys are handled."""
        col, row = key
        _tuple_err = """Tuple items must be in the format table[col, row] or
        table[row, col].
        """

        if not isinstance(col, str):
            # Try inverting
            col, row = row, col

        if not isinstance(col, str):
            raise KeyError(_tuple_err)

        if not isinstance(row, (int, slice, list, np.ndarray)):
            raise KeyError(_tuple_err)

        return col, row

    def __getitem__(self, key):
        """Get a row or a column from the table."""
        if isinstance(key, (int, np.int_)):
            return self.get_row(key)
        if isinstance(key, (str, np.str_)):
            return self.get_column(key)
        if isinstance(key, tuple):
            if len(key) not in (1, 2):
                raise KeyError(f'{key}')
            if len(key) == 1:
                return self[key[0]]
            col, row = self._resolve_tuple(key)
            return self[col][row]
        raise KeyError(f'{key}')

    def __setitem__(self, key, value):
        """Set a row or a column in the table."""
        if isinstance(key, int):
            self.set_row(key, value)
        elif isinstance(key, str):
            self.set_column(key, value)
        elif isinstance(key, tuple):
            if len(key) not in (1, 2):
                raise KeyError(f'{key}')
            if len(key) == 1:
                self[key[0]] = value
            else:
                col, row = self._resolve_tuple(key)
                self[col][row] = value
        else:
            raise KeyError(f'{key}')

    def __len__(self):
        """Get the number of rows in the table."""
        return self._db.count(self._name)

    def __contains__(self, item):
        """Check if a given column is in the table."""
        return item in self.column_names

    def __iter__(self):
        """Iterate over the rows of the table."""
        for i in self.select():
            yield i

    def __repr__(self):
        """Get a string representation of the table."""
        s = f"{self.__class__.__name__} '{self.name}'"
        s += f" in database '{self.db}':"
        s += f"({len(self.column_names)} columns x {len(self)} rows)\n"
        s += '\n'.join(self.as_table().__repr__().split('\n')[1:])
        return s


class SQLColumn:
    """Handle an SQL column operations interfacing with the DB."""

    def __init__(self, db, table, name):
        """Initialize the column.

        Parameters
        ----------
        db : SQLDatabase
            The parent database object.
        table : str
            The name of the table in the database.
        name : str
            The column name in the table.
        """
        self._db = db
        self._table = table
        self._name = name

    @property
    def name(self):
        """Get the name of the column."""
        return self._name

    @property
    def values(self):
        """Get the values of the current column."""
        vals = self._db.select(self._table, columns=[self._name])
        return [i[0] for i in vals]

    @property
    def table(self):
        """Get the table name."""
        return self._table

    def __getitem__(self, key):
        """Get a row from the column."""
        if isinstance(key, (int, np.int_, slice)):
            return self.values[key]
        if isinstance(key, (list, np.ndarray)):
            v = self.values
            return [v[i] for i in key]
        raise IndexError(f'{key}')

    def __setitem__(self, key, value):
        """Set a row in the column."""
        if isinstance(key, (int, np.int_)):
            self._db.set_item(self._table, self._name, key, value)
        elif isinstance(key, (slice, list, np.ndarray)):
            v = np.array(self.values)
            v[key] = value
            self._db.set_column(self._table, self._name, v)
        else:
            raise IndexError(f'{key}')

    def __len__(self):
        """Get the number of rows in the column."""
        return len(self.values)

    def __iter__(self):
        """Iterate over the column."""
        for i in self.values:
            yield i

    def __contains__(self, item):
        """Check if the column contains a given value."""
        return item in self.values

    def __repr__(self):
        """Get a string representation of the column."""
        s = f"{self.__class__.__name__} {self._name} in table '{self._table}'"
        s += f" ({len(self)} rows)"
        return s


class SQLRow:
    """Handle and SQL table row interfacing with the DB."""

    def __init__(self, db, table, row, colmap=None):
        """Initialize the row.

        Parameters
        ----------
        db : SQLDatabase
            The parent database object.
        table : str
            The name of the table in the database.
        row : int
            The row index in the table.
        """
        self._db = db
        self._table = table
        self._row = row
        self._colmap = colmap

    @property
    def column_names(self):
        """Get the column names of the current table."""
        names = self._db.column_names(self._table)
        if self._colmap is not None:
            names = self._colmap.get_keyword(names)
        return names

    @property
    def table(self):
        """Get the table name."""
        return self._table

    @property
    def values(self):
        """Get the values of the current row."""
        return self._db.select(self._table, where={_ID_KEY: self.index+1})[0]

    @property
    def index(self):
        """Get the index of the current row."""
        return self._row

    @property
    def keys(self):
        """Get the keys of the current row."""
        return self.column_names

    @property
    def items(self):
        """Get the items of the current row."""
        return zip(self.column_names, self.values)

    def as_dict(self):
        """Get the row as a dict."""
        return dict(self.items)

    def __getitem__(self, key):
        """Get a column from the row."""
        if isinstance(key, (str, np.str_)):
            column = key
            if self._colmap is not None:
                column = self._colmap.get_column_name(key)
            try:
                return self._db.get_item(self._table, column, self._row)
            except ValueError:
                raise KeyError(f'{key}')
        if isinstance(key, (int, np.int_)):
            return self.values[key]
        raise KeyError(f'{key}')

    def __setitem__(self, key, value):
        """Set a column in the row."""
        if not isinstance(key, (str, np.str_)):
            raise KeyError(f'{key}')

        column = key = key.lower()
        if self._colmap is not None:
            column = self._colmap.get_column_name(key)
        if key not in self.column_names:
            raise KeyError(f'{key}')
        self._db.set_item(self._table, column, self.index, value)

    def __iter__(self):
        """Iterate over the row."""
        for i in self.values:
            yield i

    def __contains__(self, item):
        """Check if the row contains a given value."""
        return item in self.values

    def __repr__(self):
        """Get a string representation of the row."""
        s = f"{self.__class__.__name__} {self._row} in table '{self._table}' "
        s += self.as_dict().__repr__()
        return s


def _sanitize_colnames(data):
    """Sanitize the colnames to avoid invalid characteres like '-'."""
    def _sanitize(key):
        if len([ch for ch in key if not ch.isalnum() and ch != '_']) != 0:
            raise ValueError(f'Invalid column name: {key}.')
        return key.lower()

    if isinstance(data, dict):
        d = data
        colnames = _sanitize_colnames(list(data.keys()))
        return dict(zip(colnames, d.values()))
    if isinstance(data, str):
        return _sanitize(data)
    if not isinstance(data, (list, tuple, np.ndarray)):
        raise TypeError(f'{type(data)} is not supported.')

    return [_sanitize(i) for i in data]


def _sanitize_value(data):
    """Sanitize the value to avoid sql errors."""
    if data is None or isinstance(data, bytes):
        return data
    if isinstance(data, (str, np.str_)):
        return f"{data}"
    if np.isscalar(data) and np.isreal(data):
        if isinstance(data, (int, np.integer)):
            return int(data)
        elif isinstance(data, (float, np.floating)):
            return float(data)
    if isinstance(data, (bool, np.bool_)):
        return bool(data)
    raise TypeError(f'{type(data)} is not supported.')


def _fix_row_index(row, length):
    """Fix the row number to be a valid index."""
    if row < 0:
        row += length
    if row >= length or row < 0:
        raise IndexError('Row index out of range.')
    return row


def _dict2row(cols, **row):
    values = [None]*len(cols)
    for i, c in enumerate(cols):
        if c in row.keys():
            values[i] = row[c]
        else:
            values[i] = None
    return values


def _parse_where(where):
    args = None
    if where is None:
        _where = None
    elif isinstance(where, dict):
        where = _sanitize_colnames(where)
        for i, (k, v) in enumerate(where.items()):
            v = _sanitize_value(v)
            if i == 0:
                _where = f"{k}=?"
                args = [v]
            else:
                _where += f" AND {k}=?"
                args.append(v)
    elif isinstance(where, str):
        _where = where
    elif isinstance(where, (list, tuple)):
        for w in where:
            if not isinstance(w, str):
                raise TypeError('if where is a list, it must be a list '
                                f'of strings. Not {type(w)}.')
        _where = ' AND '.join(where)
    else:
        raise TypeError('where must be a string, list of strings or'
                        ' dict.')
    return _where, args


class SQLDatabase:
    """Database creation and manipulation with SQL.

    Notes
    -----
    - '__id__' is only for internal indexing. It is ignored on returns.
    """

    def __init__(self, db=':memory:', autocommit=True):
        """Initialize the database.

        Parameters
        ----------
        db : str
            The name of the database file. If ':memory:' is given, the
            database will be created in memory.
        autocommit : bool (optional)
            Whether to commit changes to the database after each operation.
            Defaults to True.
        """
        self._db = db
        self._con = sql.connect(self._db)
        self._cur = self._con.cursor()
        self.autocommit = autocommit

    def execute(self, command, arguments=None):
        """Execute a SQL command in the database."""
        logger.debug('executing sql command: "%s"',
                     str.replace(command, '\n', ' '))
        try:
            if arguments is None:
                self._cur.execute(command)
            else:
                self._cur.execute(command, arguments)
            res = self._cur.fetchall()
        except sql.Error as e:
            self._con.rollback()
            raise e

        if self.autocommit:
            self.commit()
        return res

    def executemany(self, command, arguments):
        """Execute a SQL command in the database."""
        logger.debug('executing sql command: "%s"',
                     str.replace(command, '\n', ' '))

        try:
            self._cur.executemany(command, arguments)
            res = self._cur.fetchall()
        except sql.Error as e:
            self._con.rollback()
            raise e

        if self.autocommit:
            self.commit()
        return res

    def commit(self):
        """Commit the current transaction."""
        self._con.commit()

    def count(self, table, where=None):
        """Get the number of rows in the table."""
        self._check_table(table)
        comm = "SELECT COUNT(*) FROM "
        comm += f"{table} "
        where, args = _parse_where(where)
        if where is not None:
            comm += f"WHERE {where}"
        comm += ";"
        return self.execute(comm, args)[0][0]

    def select(self, table, columns=None, where=None, order=None, limit=None,
               offset=None):
        """Select rows from a table.

        Parameters
        ----------
        columns : list (optional)
            List of columns to select. If None, select all columns.
        where : dict (optional)
            Dictionary of conditions to select rows. Keys are column names,
            values are values to compare. All rows equal to the values will
            be selected. If None, all rows are selected.
        order : str (optional)
            Column name to order by.
        limit : int (optional)
            Number of rows to select.
        """
        self._check_table(table)
        if columns is None:
            columns = self[table].column_names
        elif isinstance(columns, str):
            columns = [columns]
        # only use sanitized column names
        columns = ', '.join(_sanitize_colnames(columns))

        comm = f"SELECT {columns} "
        comm += f"FROM {table} "
        args = []

        where, args_w = _parse_where(where)
        if where is not None:
            comm += f"WHERE {where} "
            if args_w is not None:
                args += args_w

        if order is not None:
            order = _sanitize_colnames(order)
            comm += f"ORDER BY {order} ASC "

        if limit is not None:
            comm += "LIMIT ? "
            if not isinstance(limit, (int, np.integer)):
                raise TypeError('limit must be an integer.')
            args.append(int(limit))
        if offset is not None:
            if limit is None:
                raise ValueError('offset cannot be used without limit.')
            if not isinstance(offset, (int, np.integer)):
                raise TypeError('offset must be an integer.')
            comm += "OFFSET ? "
            args.append(int(offset))

        comm = comm + ';'

        if args == []:
            args = None
        res = self.execute(comm, args)
        return res

    def copy(self, indexes=None):
        """Get a copy of the database."""
        return self.__copy__(indexes=indexes)

    def column_names(self, table):
        """Get the column names of the table."""
        self._check_table(table)
        comm = "SELECT * FROM "
        comm += f"{table} LIMIT 1;"
        self.execute(comm)
        return [i[0].lower() for i in self._cur.description
                if i[0].lower() != _ID_KEY.lower()]

    @property
    def db(self):
        """Get the database name."""
        return str(self._db)

    @property
    def table_names(self):
        """Get the table names in the database."""
        comm = "SELECT name FROM sqlite_master WHERE type='table';"
        return [i[0] for i in self.execute(comm) if i[0] != 'sqlite_sequence']

    def _check_table(self, table):
        """Check if the table exists in the database."""
        if table not in self.table_names:
            raise KeyError(f'Table "{table}" does not exist.')

    def _add_missing_columns(self, table, columns):
        """Add missing columns to the table."""
        existing = set(self.column_names(table))
        for col in [i for i in columns if i not in existing]:
            self.add_column(table, col)

    def _add_data_dict(self, table, data, add_columns=False,
                       skip_sanitize=False):
        """Add data sotred in a dict to the table."""
        data = _sanitize_colnames(data)
        if add_columns:
            self._add_missing_columns(table, data.keys())

        dict_row_list = _dict2row(cols=self.column_names(table), **data)
        try:
            rows = np.broadcast(*dict_row_list)
        except ValueError:
            rows = broadcast(*dict_row_list)
        rows = list(zip(*rows.iters))
        self._add_data_list(table, rows, skip_sanitize=skip_sanitize)

    def _add_data_list(self, table, data, skip_sanitize=False):
        """Add data stored in a list to the table."""
        if np.ndim(data) not in (1, 2):
            raise ValueError('data must be a 1D or 2D array.')

        if np.ndim(data) == 1:
            data = np.reshape(data, (1, len(data)))

        if np.shape(data)[1] != len(self.column_names(table)):
            raise ValueError('data must have the same number of columns as '
                             'the table.')

        if not skip_sanitize:
            data = [tuple(map(_sanitize_value, d)) for d in data]
        comm = f"INSERT INTO {table} VALUES "
        comm += f"(NULL, {', '.join(['?']*len(data[0]))})"
        comm += ';'
        self.executemany(comm, data)

    def add_table(self, table, columns=None, data=None):
        """Create a table in database."""
        logger.debug('Initializing "%s" table.', table)
        if table in self.table_names:
            raise ValueError('table {table} already exists.')

        comm = f"CREATE TABLE '{table}'"
        comm += f" (\n{_ID_KEY} INTEGER PRIMARY KEY AUTOINCREMENT"

        if columns is not None and data is not None:
            raise ValueError('cannot specify both columns and data.')
        if columns is not None:
            comm += ",\n"
            for i, name in enumerate(columns):
                comm += f"\t'{name}'"
                if i != len(columns) - 1:
                    comm += ",\n"
        comm += "\n);"

        self.execute(comm)

        if data is not None:
            self.add_rows(table, data, add_columns=True)

    def add_column(self, table, column, data=None):
        """Add a column to a table."""
        self._check_table(table)

        column = column.lower()
        if data is not None and len(data) != len(self[table]) and \
           len(self[table]) != 0:
            raise ValueError("data must have the same length as the table.")

        if column in (_ID_KEY, 'table', 'default'):
            raise ValueError(f"{column} is a protected name.")

        col = _sanitize_colnames([column])[0]
        comm = f"ALTER TABLE {table} ADD COLUMN '{col}' ;"
        logger.debug('adding column "%s" to table "%s"', col, table)
        self.execute(comm)

        # adding the data to the table
        if data is not None:
            self.set_column(table, column, data)

    def add_rows(self, table, data, add_columns=False, skip_sanitize=False):
        """Add a dict row to a table.

        Parameters
        ----------
        data : dict, list or `~numpy.ndarray`
            Data to add to the table. If dict, keys are column names,
            if list, the order of the values is the same as the order of
            the column names. If `~numpy.ndarray`, dtype names are interpreted
            as column names.
        add_columns : bool (optional)
            If True, add missing columns to the table.
        """
        self._check_table(table)
        if isinstance(data, (list, tuple)):
            return self._add_data_list(table, data,
                                       skip_sanitize=skip_sanitize)
        if isinstance(data, dict):
            return self._add_data_dict(table, data, add_columns=add_columns,
                                       skip_sanitize=skip_sanitize)
        if isinstance(data, np.ndarray):
            names = data.dtype.names
            if names is not None:
                data = {n: data[n] for n in names}
                return self._add_data_dict(table, data,
                                           add_columns=add_columns,
                                           skip_sanitize=skip_sanitize)
            return self._add_data_list(table, data,
                                       skip_sanitize=skip_sanitize)
        if isinstance(data, Table):
            data = {c: list(data[c]) for c in data.colnames}
            return self._add_data_dict(table, data, add_columns=add_columns,
                                       skip_sanitize=skip_sanitize)

        raise TypeError('data must be a dict, list, or numpy array. '
                        f'Not {type(data)}.')

    def drop_table(self, table):
        """Drop a table from the database."""
        self._check_table(table)
        comm = f"DROP TABLE {table};"
        self.execute(comm)

    def get_table(self, table, column_map=None):
        """Get a table from the database."""
        self._check_table(table)
        return SQLTable(self, table, colmap=column_map)

    def get_row(self, table, index, column_map=None):
        """Get a row from the table."""
        self._check_table(table)
        index = _fix_row_index(index, len(self[table]))
        return SQLRow(self, table, index, colmap=column_map)

    def get_column(self, table, column):
        """Get a column from the table."""
        column = column.lower()
        if column not in self.column_names(table):
            raise KeyError(f"column {column} does not exist.")
        return SQLColumn(self, table, column)

    def get_item(self, table, column, row):
        """Get an item from the table."""
        self._check_table(table)
        row = _fix_row_index(row, len(self[table]))
        column = _sanitize_colnames([column])[0]
        return self.get_column(table, column)[row]

    def set_item(self, table, column, row, value):
        """Set a value in a cell."""
        row = _fix_row_index(row, self.count(table))
        column = _sanitize_colnames([column])[0]
        value = _sanitize_value(value)
        self.execute(f"UPDATE {table} SET {column}=? "
                     f"WHERE {_ID_KEY}=?;", (value, row+1))

    def set_row(self, table, row, data):
        """Set a row in the table."""
        row = _fix_row_index(row, self.count(table))
        colnames = self.column_names(table)

        if isinstance(data, dict):
            data = _dict2row(colnames, **data)
        elif isinstance(data, (list, tuple, np.ndarray)):
            if len(data) != len(colnames):
                raise ValueError('data must have the same length as the '
                                 'table.')
        else:
            raise TypeError('data must be a dict, list, or numpy array. '
                            f'Not {type(data)}.')

        comm = f"UPDATE {table} SET "
        comm += f"{', '.join(f'{i}=?' for i in colnames)} "
        comm += f" WHERE {_ID_KEY}=?;"
        self.execute(comm, tuple(list(map(_sanitize_value, data)) + [row+1]))

    def set_column(self, table, column, data):
        """Set a column in the table."""
        tablen = self.count(table)
        if column not in self.column_names(table):
            raise KeyError(f"column {column} does not exist.")
        if len(data) != tablen and tablen != 0:
            raise ValueError("data must have the same length as the table.")

        if tablen == 0:
            for i in range(len(data)):
                self.add_rows(table, {})

        col = _sanitize_colnames([column])[0]
        comm = f"UPDATE {table} SET "
        comm += f"{col}=? "
        comm += f" WHERE {_ID_KEY}=?;"
        args = list(zip([_sanitize_value(d) for d in data],
                        range(1, self.count(table)+1)))
        self.executemany(comm, args)

    def index_of(self, table, where):
        """Get the index(es) where a given condition is satisfied."""
        indx = self.select(table, _ID_KEY, where=where)
        if len(indx) == 1:
            return indx[0][0]-1
        return [i[0]-1 for i in indx]

    def __len__(self):
        """Get the number of rows in the current table."""
        return len(self.table_names)

    def __del__(self):
        """Delete the class, closing the db connection."""
        # ensure connection is closed.
        self._con.close()

    def __setitem__(self, item, value):
        """Set a row in the table."""
        if not isinstance(item, tuple):
            raise KeyError('item must be a in the formats '
                           'db[table, row], db[table, column] or '
                           'db[table, column, row].')
        if not isinstance(item[0], str):
            raise KeyError('first item must be the table name.')
        self.get_table(item[0])[item[1:]] = value

    def __getitem__(self, item):
        """Get a items from the table."""
        if isinstance(item, (str, np.str_)):
            return self.get_table(item)
        if isinstance(item, tuple):
            if not isinstance(item[0], str):
                raise ValueError('first item must be the table name.')
            return self.get_table(item[0])[item[1:]]
        raise ValueError('items must be a string for table names, '
                         'or a tuple in the formats. '
                         'db[table, row], db[table, column] or '
                         'db[table, column, row].')

    def __repr__(self):
        """Get a string representation of the table."""
        s = f"{self.__class__.__name__} '{self.db}' at {hex(id(self))}:"
        if len(self) == 0:
            s += '\n\tEmpty database.'
        for i in self.table_names:
            s += f"\n\t{i}: {len(self.column_names(i))} columns"
            s += f" {len(self[i])} rows"
        return s

    def __copy__(self, indexes=None):
        """Copy the database.

        Parameters
        ----------
        indexes : dict, optional
            A dictionary of table names and their indexes to copy.

        Returns
        -------
        db : SQLDatabase
            A copy of the database.
        """
        def _get_data(table, indx=None):
            if indx is None:
                return self.select(table)
            if len(indx) == 0:
                return None
            if len(indx) >= 10000:
                # too many rows must be divided
                indx = np.array_split(indx, np.ceil(len(indx)/10000))
                d = None
                for i in indx:
                    if d is None:
                        d = _get_data(table, i)
                    else:
                        np.vstack([d, _get_data(table, i)])
                return d
            where = " OR ".join(f"{_ID_KEY}={v+1}" for v in indx)
            return self.select(table, where=where)

        # when copying, always copy to memory
        db = SQLDatabase(':memory:')
        if indexes is None:
            indexes = {}
        for i in self.table_names:
            db.add_table(i, columns=self.column_names(i))
            rows = _get_data(i, indexes.get(i, None))
            if rows is not None:
                db.add_rows(i, rows, skip_sanitize=True)
        return db



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import sep
import numpy as np
from scipy.stats import sigmaclip
from astropy.table import Table

from ..math.array import xy2r, trim_array
from ._utils import _sep_fix_byte_order
from .detection import calc_fwhm
from ..logger import logger
from ..fits_utils import imhdus


def sky_annulus(data, x, y, r_ann, algorithm='mmm', mask=None):
    """Determine the sky value of a single pixel based on a sky annulus
       and obtains the sky value by calculating the median of annulus pixels
       or by the MMM mode estimator from DAOPHOT (3*median - 2*mean).

    Parameters
    ----------
    data : `~nnumpy.ndarray`
        2D image data for photometry
    x, y : array_like
        Positions of the sources
    r_ann : array_like([float, float])
        Annulus radius (intern and extern) to calculate the background
        value
    algorithm : 'mmm' or 'sigmaclip' (optional)
        Algorith to calculate the background value. 'mmm' (mean, median,
        mode) should be better for populated fields, while 'sigmaclip'
        (clipped mean) should be better for sparse fields.
        Default: 'mmm'

    Returns
    -------
    sky : array_like
        The computed value of sky for each (x, y) source.
    sky_error : array_like
        The error of sky value, computed as the sigma cliped stddev.
    """
    # TODO: this code needs optimization for faster work
    if len(x) != len(y):
        raise ValueError('x and y variables don\'t have the same lenght.')

    if len(r_ann) != 2:
        raise ValueError('r_ann must have two components (r_in, r_out)')

    sky = np.zeros_like(x, dtype='f8')
    sky.fill(np.nan)
    sky_error = np.zeros_like(x, dtype='f8')
    sky_error.fill(np.nan)

    box_size = 2*int(np.max(r_ann)+2)  # Ensure the aperture is entirely in box
    r_ann = sorted(r_ann)

    indices = np.indices(data.shape)
    for i in range(len(x)):
        xi, yi = x[i]-0.5, y[i]-0.5  # needed to check pixel centers
        d, ix, iy = trim_array(data, box_size, (xi, yi), indices)
        if mask is not None:
            m = np.ravel(mask[iy, ix])
        r, f = xy2r(ix, iy, d, xi, yi)

        # Filter only values inside the annulus
        # To think: this do not perform subpixel, just check the pixel center
        filt = (r >= r_ann[0]) & (r <= r_ann[1])
        # mask nans here to go faster
        filt = filt & ~np.isnan(f)
        if mask is not None:
            filt = filt & ~m
        f = f[np.where(filt)]
        if len(f) < 1:
            logger.warn('No pixels for sky subtraction found at position'
                        f' {x[i]}x{y[i]}.')
            sky[i] = 0
            sky_error[i] = 0
        else:
            for _ in range(3):
                f, _, _ = sigmaclip(f)
            mean = np.nanmean(f)
            median = np.nanmedian(f)
            sky_error[i] = np.nanstd(f)
            if algorithm == 'mmm':
                # mimic daophot using sigmaclip
                sky[i] = 3*median - 2*mean  # mmm mode estimator
            elif algorithm == 'sigmaclip':
                sky[i] = mean
            else:
                raise ValueError(f'Sky algorithm {algorithm} not supported.')

    return sky, sky_error


def aperture_photometry(data, x, y, r='auto', r_ann='auto', gain=1.0,
                        readnoise=None, mask=None, sky_algorithm='mmm'):
    """Perform aperture photometry using sep.

    Parameters
    ----------
    data : np.ndarray
        2D image data for photometry
    x, y : array_like
        Positions of the sources
    r : float or 'auto' (optional)
        Aperture radius. If `auto`, the value will be estimated based in
        the median gaussian FWHM of the sources in the image
        (r=0.6731*GFWHM).
        Default: 'auto'
    r_ann : array_like([float, float]), None or 'auto' (optional)
        Annulus radii (r_in, r_out) for local background extraction.
        If `auto`, the annulus will be set based on aperture as (4*r, 6*r).
        If None, no local background subtraction will be performed.
        Default: 'auto'
    gain : float (optional)
        Gain to correctly calculate the error.
    readnoise : float (optional)
        Readnoise of the image to correctly calculate the error.
    mask : np.ndarray (optional)
        Mask badpixels and problematic ccd areas.
    sky_algorithm : 'mmm' or 'sigmaclip'
        Algorith to calculate the background value. 'mmm' (mean, median,
        mode) should be better for populated fields, while 'sigmaclip'
        (clipped median) should be better for sparse fields.
        Default: 'mmm'

    Returns
    -------
    res_ap : `~astropy.table.Table`
        Table containing all aperture photometry informations.
        - ``x``, ``y``: position of the sources
        - ``aperture``: aperture radius
        - ``flux``: flux of the sources
        - ``flux_error``: flux error of the sources
        - ``sky``: sky background value
        - ``sky_error``: sky background error
        - ``flag``: flag for the sources
    """
    res_ap = Table()

    if isinstance(data, imhdus):
        data = data.data

    data = _sep_fix_byte_order(data)
    x = np.array(x)
    y = np.array(y)

    if gain is None:
        gain = 1.0
    kwargs = {'gain': gain,
              'err': readnoise,
              'subpix': 0}

    res_ap['x'] = x
    res_ap['y'] = y

    if r == 'auto':
        logger.debug('Aperture r set as `auto`. Calculating from FWHM.')
        fwhm = calc_fwhm(data, x, y, box_size=25, model='gaussian')
        r = 0.6371*fwhm
        res_ap.meta['fwhm'] = fwhm
        res_ap.meta['r_auto'] = True
        logger.debug(f'FWHM:{fwhm} r:{r}')

    res_ap['aperture'] = [r]*len(x)

    if r_ann == 'auto':
        logger.debug('Aperture r_ann set as `auto`. Calculating from r.')
        r_in = int(round(4*r, 0))
        r_out = int(max(r_in+10, round(6*r, 0)))  # Ensure a dannulus geq 10
        r_ann = (r_in, r_out)
        logger.debug(f'r_ann:{r_ann}')

    sky = None

    flux, flux_error, flags = sep.sum_circle(data, x, y, r, mask=mask,
                                             **kwargs)
    if r_ann is not None:
        ri, ro = r_ann
        res_ap.meta['r_in'] = ri
        res_ap.meta['r_out'] = ro

        sky, error = sky_annulus(data, x, y, r_ann, algorithm=sky_algorithm,
                                 mask=mask)

        # SEP do not expose aperture area, so we calculate
        nkwargs = kwargs.copy()
        nkwargs['gain'] = 1.0  # ensure gain not change the area
        ones = _sep_fix_byte_order(np.ones(data.shape, dtype='<f8'))
        area, _, _ = sep.sum_circle(ones, x, y, r, mask=mask, **nkwargs)

        # TODO: check these calculations
        # recompute flux, flux_error and flags
        flux -= sky*area
        bkgerr = np.sqrt((error + np.absolute(sky)/gain)*area)
        flux_error = np.hypot(flux_error, bkgerr)

    res_ap['flux'] = flux
    res_ap['flux_error'] = flux_error
    res_ap['flags'] = flags

    if sky is not None:
        res_ap['sky'] = sky

    return res_ap



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Small python addons to be used in astropop."""

import subprocess
import shlex
from numbers import Number

import numpy as np

from .logger import logger, resolve_level_string

__all__ = ['string_fix', 'process_list', 'check_iterable',
           'batch_key_replace', 'IndexedDict', 'check_number',
           'broadcast']


def process_list(func, iterator, *args, **kwargs):
    """Run a function func for all i in a iterator list.

    The processing will occurr in serial form. The function will be
    processed multiple times using the values iterated from the `iterator`.

    Parameters
    ----------
    func: callable
        Function to be executed with the iterator argument.
    iterator: list
        Values to be passed to the func.
    *args, **kwargs:
        Other, non-varying, arguments to be passed to the function.

    Returns
    -------
    list:
        A list with the results.
    """
    return [func(i, *args, **kwargs) for i in iterator]


def string_fix(string, encode='utf-8'):
    """
    Fix the byte<->string problem in python 3.

    This method converts anything is passed to it into a string. If it is
    `bytes` also handle the decoding.

    Parameters
    ----------
    string: `str` or `bytes`
        String to be checked or converted.
    encode: string (optional)
        Python compatible `bytes` like encoding.
        Default: 'utf-8'

    Returns
    -------
    string: `str`
        The Python 3 string.
    """
    if isinstance(string, bytes):
        string = string.decode(encode)
    elif not isinstance(string, str):
        # everything is not bytes or string
        string = str(string)

    return string


def check_iterable(value):
    """Check if a value is iterable (list), but not a string.

    The checking process is done by trying the `iter` method of Python and
    matching if the value is a string or not.

    Parametes
    ---------
    value: any
        Value to be checked if it is iterable.

    Returns
    -------
    bool:
        `True` if the value is iterable and not a string. `False` otherwise.
    """
    try:
        iter(value)
        if isinstance(value, (str, bytes)):
            return False
        return True
    except NotImplementedError:
        # multi-dimensional sub-views are not implemented
        if isinstance(value, np.memmap):
            return True
    except TypeError:
        pass

    return False


def check_number(value):
    """Check if a value passed is a scalar number.

    Parameters
    ----------
    value: any
        Value to be checked.

    Returns
    -------
    bool:
        `True` if the value is a scalar number (not string, list or array) and
        `False` otherwise.
    """
    if isinstance(value, Number):
        if not isinstance(value, (bool, complex)):
            return True
    return False


class _scalar_iterator:
    """Iterator for scalar values."""
    def __init__(self, value, length):
        self.value = value
        self.length = length

    def __iter__(self):
        for i in range(self.length):
            yield self.value

    def __getitem__(self, index):
        if index >= self.length or index < 0:
            raise IndexError
        return self.value

    def __len__(self):
        return self.length


class broadcast:
    """Broadcast values to the list size. Alternative to `~numpy.broadcast`."""

    def __init__(self, *args):
        """Initialize the broadcast object."""
        if len(args) == 0:
            raise ValueError("Empty broadcast")

        lengths = np.array([self._get_length(a) for a in args])
        if np.all(lengths == -1):
            self.length = 1
        else:
            self.length = [i for i in lengths if i >= 0]
            if len(set(self.length)) > 1:
                raise ValueError("All array arguments must have the same "
                                 "length.")
            self.length = self.length[0]
        self.args = [a if lengths[i] >= 0 else _scalar_iterator(a, self.length)
                     for i, a in enumerate(args)]

    def __iter__(self):
        """Return the iterator."""
        for i in zip(*self.iters):
            yield i

    @staticmethod
    def _get_length(value):
        """Get the length of iterable only values."""
        if not check_iterable(value):
            return -1
        return len(value)

    @property
    def iters(self):
        """Return the tuple containing the iterators."""
        return self.args

    def __len__(self):
        """Return the length of the broadcast."""
        return self.length


def batch_key_replace(dictionary, key=None):
    """Scan and replace {key} values in a dict by dict['key'] value.

    Notes
    -----
    All the replacement is did inplace. Nothing is returned.

    Parameters
    ----------
    dictionary: dict_like
        Dictionary to replace the keys.
    key: string or None, optional
        The key to be replaced.
    """
    if key is None:
        for i in dictionary.keys():
            batch_key_replace(dictionary, i)
        return

    if isinstance(dictionary[key], (str, bytes)):
        for i in dictionary.keys():
            if '{'+i+'}' in dictionary[key]:
                logger.debug("replacing key %s in key %s", i, key)
                batch_key_replace(dictionary, i)
                dictionary[key] = dictionary[key].format(**{i: dictionary[i]})
    elif check_iterable(dictionary[key]):
        for j in range(len(dictionary[key])):
            for i in dictionary.keys():
                v = dictionary[key][j]
                if '{'+i+'}' in str(v):
                    logger.debug("replacing key %s in key %s", i, key)
                    dictionary[key][j] = v.format(**{i: dictionary[i]})
    else:
        return


def run_command(args, stdout=None, stderr=None, stdout_loglevel='DEBUG',
                stderr_loglevel='ERROR', **kwargs):
    """Run a command in command line with logging."""

    # Put the cmd in python list, required
    if isinstance(args, (str, bytes)):
        logger.debug('Converting string using shlex')
        args = shlex.split(args)

    ps = {
        'out': {'log': resolve_level_string(stdout_loglevel), 'list': stdout},
        'err': {'log': resolve_level_string(stderr_loglevel), 'list': stderr}
    }

    logger.log(ps['out']['log'], 'Runing: %s', " ".join(args))
    # Run the command
    proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            **kwargs)
    while True:
        output = proc.stdout.readline().decode('utf-8').strip()
        err = proc.stderr.readline().decode('utf-8').strip()
        if output == '' and err == '' and proc.poll() is not None:
            break
        if output:
            logger.log(ps['out']['log'], output)
            if ps['out']['list'] is not None:
                ps['out']['list'].append(output)
        if err:
            logger.log(ps['err']['log'], err)
            if ps['err']['list'] is not None:
                ps['err']['list'].append(err)

    logger.log(ps['out']['log'], "Done with process: %s", " ".join(args))

    return proc, stdout, stderr


class IndexedDict(dict):
    """Extends Python dictionary to include indexing and inserting.

    Python3.7 keeps assignment ordering in default dict, like OrderedDict.
    """

    def index(self, key):
        """Return the index of a key in the list."""
        __keys = self.keys()

        if key not in __keys:
            raise KeyError(f"{key}")

        return np.where(__keys == key)[0]

    def insert_before(self, key, new_key, val):
        """Insert new_key:value into dict before key."""
        index = self.index(key)
        self.insert_at(index, new_key, val)

    def insert_after(self, key, new_key, val):
        """Insert new_key:value into dict after key."""
        index = self.index(key)
        self.insert_at(index+1, new_key, val)

    def insert_at(self, index, key, value):
        """Insert a key:value to an specific index."""
        __keys = list(self.keys())
        __vals = list(self.values())

        # Romeve existing keys
        if key in __keys:
            ind = __keys.index(key)
            __keys.pop(ind)
            __vals.pop(ind)
            if index > ind:
                index = index-1

        if index < (len(__keys) - 1):
            __keys.insert(index, key)
            __vals.insert(index, value)
            self.clear()
            self.update({x: __vals[i] for i, x in enumerate(__keys)})
        else:
            self.update({key: value})



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import hashlib


__all__ = ['hasher']


def hasher(string, size=8):
    """Simple function to generate a SHA1 hash of a string.

    Parameters:
        - string : string or bytes
            The string to be hashed.
        - size : int
            Size of the output hash string.

    Returns:
        - h : string
            Hash string trunked to size.
    """
    string = str(string)
    h = hashlib.sha256(string.encode()).hexdigest()
    return h[:size]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Moffat models and utilities."""

import numpy as np
from astropy.modeling import Fittable1DModel, Fittable2DModel, Parameter
# from scipy.special import gamma as G


__all__ = ['moffat_r', 'moffat_1d', 'moffat_2d', 'PSFMoffat2D',
           'PSFMoffat1D', 'MoffatRadial']


def moffat_bounding_box(alpha, beta, threshold=10**-3):
    """Calculate the radius where the distribution reaches a given threshold.

    Threshold relative to maximum.
    """
    return int(np.sqrt((threshold**(-1/beta) - 1)*alpha))


def moffat_normalize(width, power):
    # From Yoonsoo Bach notebook
    # http://nbviewer.jupyter.org/github/ysbach/AO_2017/blob/master/04_Ground_Based_Concept.ipynb
    # return G(power)/(width*np.sqrt(np.pi)*G(power - 1/2))
    # Looking to the wikipedia, the PDF of Moffat distribution gives:
    return (power - 1)/(np.pi*width**2)


def moffat_fwhm(alpha, beta):
    return np.abs(2*alpha*np.sqrt(2**(1/beta) - 1))


def moffat_r(r, alpha, beta, flux, sky):
    sf = flux*moffat_normalize(alpha, beta)
    return sky + sf*(1+(r/alpha)**2)**(-beta)


def moffat_1d(x, x0, alpha, beta, flux, sky):
    r = x-x0
    return moffat_r(r, alpha, beta, flux, sky)


def moffat_2d(x, y, x0, y0, alpha, beta, flux, sky):
    # TODO: make this function assymetrical
    r = np.hypot(x-x0, y-y0)
    return moffat_r(r, alpha, beta, flux, sky)


class PSFMoffat2D(Fittable2DModel):
    flux = Parameter(default=1, fixed=False)
    x_0 = Parameter(default=0)
    y_0 = Parameter(default=0)
    alpha = Parameter(default=1, fixed=False)
    beta = Parameter(default=1.5, fixed=False)
    sky = Parameter(default=0, fixed=False)

    @staticmethod
    def evaluate(x, y, x_0, y_0, flux, alpha, beta, sky):
        return moffat_2d(x, y, x_0, y_0, alpha, beta, flux, sky)

    @property
    def fwhm(self):
        return moffat_fwhm(self.alpha, self.beta)


class PSFMoffat1D(Fittable1DModel):
    flux = Parameter(default=1, fixed=False)
    x_0 = Parameter(default=0)
    alpha = Parameter(default=1, fixed=False)
    beta = Parameter(default=1.5, fixed=False)
    sky = Parameter(default=0, fixed=False)

    @staticmethod
    def evaluate(x, x_0, flux, alpha, beta, sky):
        return moffat_1d(x, x_0, alpha, beta, flux, sky)

    @property
    def fwhm(self):
        return moffat_fwhm(self.alpha, self.beta)


class MoffatRadial(Fittable1DModel):
    flux = Parameter(default=1, fixed=False)
    alpha = Parameter(default=1, fixed=False)
    beta = Parameter(default=1.5, fixed=False)
    sky = Parameter(default=0, fixed=False)

    @staticmethod
    def evaluate(x, flux, alpha, beta, sky):
        return moffat_r(x, alpha, beta, flux, sky)

    @property
    def fwhm(self):
        return moffat_fwhm(self.alpha, self.beta)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import os
import numpy as np
import pytest

from astropy.io import fits

from astropy.utils import NumpyRNGContext
from astropop.framedata import FrameData
from astropop.logger import logger, log_to_list
from astropop.image.imcombine import imcombine, _sigma_clip, \
                                     _minmax_clip, ImCombiner
from astropop.testing import *


class Test_MinMaxClip():
    def test_1D_simple(self):
        arr = np.arange(10)
        low, high = (2, 6)
        expect = np.array([1, 1, 0, 0, 0, 0, 0, 1, 1, 1], dtype=bool)

        mask = _minmax_clip(arr, low, high)
        assert_equal(mask, expect)

    def test_2D_simple(self):
        arr = np.arange(10).reshape((2, 5))
        low, high = (2, 6)
        expect = np.array([1, 1, 0, 0, 0, 0, 0, 1, 1, 1],
                          dtype=bool).reshape((2, 5))

        mask = _minmax_clip(arr, low, high)
        assert_equal(mask, expect)

    def test_3D_simple(self):
        arr = np.array([[[0, 1, 1], [2, 3, 3], [1, 2, 3]],
                        [[2, 3, 4], [2, 5, 6], [1, 0, 0]],
                        [[0, 1, 1], [2, 3, 7], [7, 0, 1]]])
        low, high = (2, 6)
        exp = np.array([[[1, 1, 1], [0, 0, 0], [1, 0, 0]],
                        [[0, 0, 0], [0, 0, 0], [1, 1, 1]],
                        [[1, 1, 1], [0, 0, 1], [1, 1, 1]]], dtype=bool)

        mask = _minmax_clip(arr, low, high)
        assert_equal(mask, exp)

    def test_minimum_disabled(self):
        arr = np.array([0, 1, 0, 2, 3, 0])
        low, high = (None, 1)
        expect = np.array([0, 0, 0, 1, 1, 0])
        mask = _minmax_clip(arr, low, high)
        assert_equal(mask, expect)

    def test_maximum_disabled(self):
        arr = np.array([0, 1, 0, 2, 3, 0])
        low, high = (1, None)
        expect = np.array([1, 0, 1, 0, 0, 1])
        mask = _minmax_clip(arr, low, high)
        assert_equal(mask, expect)

    def test_invalid(self):
        # must mask invalid numbers
        arr = np.array([0, 1, 2, np.inf, np.nan, 5, 1])
        low, high = (1, 3)
        expect = np.array([1, 0, 0, 1, 1, 1, 0])
        mask = _minmax_clip(arr, low, high)
        assert_equal(mask, expect)


class Test_SigmaClip():
    # TODO: test 3D
    # TODO: test axis in 3D
    def test_unkown_sigmaclip(self):
        arr = np.arange(10)
        with pytest.raises(TypeError):
            _sigma_clip(arr, 'testing')

        with pytest.raises(TypeError):
            _sigma_clip(arr, '1,2')

        # too many values must raise
        with pytest.raises(ValueError):
            _sigma_clip(arr, [1, 2, 3])

    def test_sigmclip_types(self):
        arr = np.arange(10)

        # must work with numbers
        _sigma_clip(arr, 1)

        # must work with 2-elements array
        _sigma_clip(arr, np.array([1, 2]))
        _sigma_clip(arr, [1, 2])
        _sigma_clip(arr, (1, 2))

    def test_invalid(self):
        arr = np.ones((5, 5))
        exp = np.zeros((5, 5), dtype=bool)

        indx = [(1, 1), (2, 1), (2, 3)]

        arr[indx[0]] = 1000
        arr[indx[1]] = np.inf
        arr[indx[2]] = np.nan

        for i in indx:
            exp[i] = True

        mask = _sigma_clip(arr)
        assert_equal(mask, exp)

    def test_functions_names(self):
        arr = np.ones((5, 5))

        # all this should run
        _sigma_clip(arr, cen_func='median')
        _sigma_clip(arr, cen_func='mean')
        _sigma_clip(arr, dev_func='std')
        _sigma_clip(arr, cen_func='mean', dev_func='std')
        _sigma_clip(arr, dev_func='mad_std')
        _sigma_clip(arr, cen_func='median', dev_func='mad_std')

    def test_functions_callable(self):
        arr = np.ones((5, 5))

        mask = _sigma_clip(arr, cen_func=np.median)
        mask = _sigma_clip(arr, dev_func=np.std)

        # testing forced 1.0pm0.5
        def _test_cen(*args, **kwargs):
            return 1.0

        def _test_dev(*args, **kwargs):
            return 0.5

        # 1.2 should not be masked with 1.0pm0.5, 2.0 and 1000 yes
        arr[0, 0] = 1.2
        arr[1, 1] = 2
        arr[3, 2] = 1000

        mask = _sigma_clip(arr, 1, cen_func=_test_cen, dev_func=_test_dev)
        for i in range(5):
            for j in range(5):
                if (i, j) in [(1, 1), (3, 2)]:
                    assert_true(mask[i, j])
                else:
                    assert_false(mask[i, j])

    def test_functions_invalid(self):
        arr = np.ones((5, 5))

        with pytest.raises(KeyError, match='invalid'):
            _sigma_clip(arr, cen_func='invalid')
        with pytest.raises(KeyError, match='invalid'):
            _sigma_clip(arr, dev_func='invalid')
        with pytest.raises(KeyError, match='1'):
            _sigma_clip(arr, cen_func=1)
        with pytest.raises(KeyError, match='1'):
            _sigma_clip(arr, dev_func=1)
        with pytest.raises(TypeError):
            _sigma_clip(arr, cen_func=[])
        with pytest.raises(TypeError):
            _sigma_clip(arr, dev_func=[])

    def test_1D_simple(self):
        with NumpyRNGContext(123):
            arr = np.random.normal(5, 2, 100)
        arr[32] = 1000
        arr[25] = 1000
        arr[12] = -1000
        exp = np.zeros(100, dtype=bool)
        exp[32] = True
        exp[25] = True
        exp[12] = True

        mask = _sigma_clip(arr, 3)
        assert_equal(mask, exp)

    def test_2D_defaults(self):
        # mean:4.532 median:4.847 std:282 mad_std: 2.02
        arr = np.array([[6.037, 5.972, 5.841, 2.775, 0.711],
                        [6.539, 4.677, -1000, 5.633, 3.478],
                        [4.847, 7.563, 3.323, 7.952, 6.646],
                        [6.136, 2.690, 4.193, 1000., 4.483],
                        [5.673, 7.479, 3.874, 4.756, 2.021]])
        # using defaults median and mad_std
        expect_1 = [[0, 0, 0, 1, 1],
                    [0, 0, 1, 0, 0],
                    [0, 1, 0, 1, 0],
                    [0, 1, 0, 1, 0],
                    [0, 1, 0, 0, 1]]
        expect_1 = np.array(expect_1, dtype=bool)

        expect_3 = [[0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0]]
        expect_3 = np.array(expect_3, dtype=bool)

        mask = _sigma_clip(arr, 3)
        assert_equal(mask, expect_3)


class Test_ImCombineConformance():
    def test_class_creation(self):
        # defaults
        c = ImCombiner()
        assert_is_instance(c, ImCombiner)
        assert_equal(c._max_memory, 1e9)
        assert_equal(c._dtype, np.float64)

        d = ImCombiner(max_memory=2e10)
        assert_equal(d._max_memory, 2e10)

        e = ImCombiner(dtype=np.float16)
        assert_equal(e._dtype, np.float16)

        f = ImCombiner(dtype=np.float32)
        assert_equal(f._dtype, np.float32)

    def test_creation_error_dtype(self):
        msg = "Only float dtypes are allowed in ImCombiner."
        with pytest.raises(ValueError, match=msg):
            ImCombiner(dtype=np.int)
        with pytest.raises(ValueError, match=msg):
            ImCombiner(dtype=np.int16)

        with pytest.raises(ValueError, match=msg):
            ImCombiner(dtype=np.int32)

        with pytest.raises(ValueError, match=msg):
            ImCombiner(dtype=np.complex64)

        with pytest.raises(ValueError, match=msg):
            ImCombiner(dtype=np.complex128)

    def test_creation_default_dtypes(self):
        c = ImCombiner(dtype=float)
        assert_equal(c._dtype, float)

        msg = "Only float dtypes are allowed in ImCombiner."
        with pytest.raises(ValueError, match=msg):
            ImCombiner(dtype=int)

        with pytest.raises(ValueError, match=msg):
            ImCombiner(dtype=complex)

    def test_class_set_sigmaclip(self):
        # empty must disable clipping
        c = ImCombiner()
        c.set_sigma_clip()
        assert_equal(c._sigma_clip, None)
        assert_equal(c._sigma_cen_func, None)
        assert_equal(c._sigma_dev_func, None)

        # set sigmaclip one value
        c = ImCombiner()
        c.set_sigma_clip(1)
        assert_equal(c._sigma_clip, 1)
        assert_equal(c._sigma_cen_func, 'median')
        assert_equal(c._sigma_dev_func, 'mad_std')

        # set sigmaclip two values
        c = ImCombiner()
        c.set_sigma_clip((1, 2))
        assert_equal(c._sigma_clip, (1, 2))
        assert_equal(c._sigma_cen_func, 'median')
        assert_equal(c._sigma_dev_func, 'mad_std')

        # Enable and disable
        c = ImCombiner()
        c.set_sigma_clip((1, 2))
        c.set_sigma_clip()
        assert_equal(c._sigma_clip, None)
        assert_equal(c._sigma_cen_func, None)
        assert_equal(c._sigma_dev_func, None)

        # set functions
        c = ImCombiner()
        c.set_sigma_clip((1, 2), 'mean', 'mad_std')
        assert_equal(c._sigma_clip, (1, 2))
        assert_equal(c._sigma_cen_func, 'mean')
        assert_equal(c._sigma_dev_func, 'mad_std')

    def test_class_set_sigmaclip_errors(self):
        # more then 2 elements must fail
        with pytest.raises(ValueError):
            c = ImCombiner()
            c.set_sigma_clip((1, 2, 3))

        # problematic functions
        with pytest.raises(ValueError):
            c = ImCombiner()
            c.set_sigma_clip((1, 2), 'no-existing')

        with pytest.raises(ValueError):
            c = ImCombiner()
            c.set_sigma_clip((1, 2), 'mean', 'no-existing')

        # None should fail
        with pytest.raises(ValueError):
            c = ImCombiner()
            c.set_sigma_clip((1, 2), None)

        with pytest.raises(ValueError):
            c = ImCombiner()
            c.set_sigma_clip((1, 2), 'mean', None)

    def test_class_set_minmax(self):
        c = ImCombiner()
        c.set_minmax_clip(0, 1)
        assert_equal(c._minmax, (0, 1))

        # test flipping
        c = ImCombiner()
        c.set_minmax_clip(1, 0)
        assert_equal(c._minmax, (0, 1))

        # test nones
        c = ImCombiner()
        c.set_minmax_clip(None, 0)
        assert_equal(c._minmax, (None, 0))
        c = ImCombiner()
        c.set_minmax_clip(0, None)
        assert_equal(c._minmax, (0, None))

        # test one element setting
        c = ImCombiner()
        c.set_minmax_clip(0)
        assert_equal(c._minmax, (0, None))
        c = ImCombiner()
        c.set_minmax_clip(min_value=0)
        assert_equal(c._minmax, (0, None))
        c = ImCombiner()
        c.set_minmax_clip(max_value=0)
        assert_equal(c._minmax, (None, 0))

        # disable
        c = ImCombiner()
        c.set_minmax_clip()
        assert_equal(c._minmax, None)
        c.set_minmax_clip(0, 1)
        assert_equal(c._minmax, (0, 1))
        c.set_minmax_clip()
        assert_equal(c._minmax, None)

    def test_class_set_minmax_errors(self):
        c = ImCombiner()

        # not numbers should fail
        with pytest.raises(ValueError):
            c.set_minmax_clip('a')

        # not numbers should fail
        with pytest.raises(ValueError):
            c.set_minmax_clip(0, 'a')

        # not numbers should fail
        with pytest.raises(ValueError):
            c.set_minmax_clip([1, 2])

        # not numbers should fail
        with pytest.raises(ValueError):
            c.set_minmax_clip(0, [1, 2])

    def test_check_consistency(self):
        n = 10
        d = np.ones((10, 10))
        l = [FrameData(d, unit='adu') for i in range(n)]
        comb = ImCombiner()
        # empty should raise
        with pytest.raises(ValueError, match='Combiner have no images.'):
            comb._check_consistency()

        comb._load_images(l)
        # nothing should raise
        comb._check_consistency()

        # incompatible unit should raise
        comb._images[3].unit = 'm'
        with pytest.raises(ValueError, match='.* unit incompatible .*'):
            comb._check_consistency()
        comb._images[3].unit = 'adu'

        # incompatible shape should raise
        comb._images[4].data = np.ones((2, 2))
        with pytest.raises(ValueError, match='.* shape incompatible .*'):
            comb._check_consistency()

    def test_invalid_method(self):
        n = 10
        d = np.ones((10, 10))
        l = [FrameData(d, unit='adu') for i in range(n)]
        comb = ImCombiner()
        with pytest.raises(ValueError, match='hulk-smash is not a valid '
                           'combining method.'):
            comb.combine(l, method='hulk-smash')


class Test_ImCombiner_ChunkYielder():
    @pytest.mark.parametrize('method', ['mean', 'median', 'sum'])
    def test_chunk_yielder_f64(self, method):
        n = 100
        d = np.random.random((100, 100)).astype(np.float64)
        l = [FrameData(d, unit='adu') for i in range(n)]
        # data size = 8 000 000 = 8 bytes * 100 * 100 * 100
        # mask size = 1 000 000 = 1 bytes * 100 * 100 * 100
        # total size = 9 000 000

        comb = ImCombiner(max_memory=1e6, dtype=np.float64)
        comb._load_images(l)

        logs = []
        lh = log_to_list(logger, logs, False)
        level = logger.getEffectiveLevel()
        logger.setLevel('DEBUG')

        if method == 'median':
            # for median, tot_size=9*4.5=41
            # xstep = 2, so n_chuks=50
            nchunk = 50
            shape = [(n, 2, 100)]
        elif method in ['mean', 'sum']:
            # for mean and sum, tot_size=9*3=27
            # xstep = 3, so n_chunks=33+1
            nchunk = 34
            shape = [(n, 3, 100), (n, 1, 100)]

        i = 0
        for chunk, unct, slc in comb._chunk_yielder(method=method):
            i += 1
            assert_in(chunk.shape, shape)
            assert_equal(chunk[0], d[slc])
            assert_is_none(unct)
            assert_is_instance(chunk, np.ndarray)
        assert_equal(i, nchunk)
        assert_in(f'Splitting the images into {nchunk} chunks.', logs)
        logs.clear()

        # this should not split into chunks
        comb = ImCombiner(max_memory=1e8)
        comb._load_images(l)

        i = 0
        for chunk, unct, slc in comb._chunk_yielder(method=method):
            i += 1
            assert_equal(chunk.shape, (n, 100, 100))
            assert_equal(chunk[0], d)
            assert_is_none(unct)
            assert_is_instance(chunk, np.ndarray)
        assert_equal(i, 1)

    @pytest.mark.parametrize('method', ['mean', 'median', 'sum'])
    def test_chunk_yielder_f32(self, method):
        # using float32, the number of chunks are almost halved
        n = 100
        d = np.random.random((100, 100)).astype(np.float64)
        l = [FrameData(d, unit='adu') for i in range(n)]
        # data size = 4 000 000 = 4 bytes * 100 * 100 * 100
        # mask size = 1 000 000 = 1 bytes * 100 * 100 * 100
        # total size = 5 000 000

        comb = ImCombiner(max_memory=1e6, dtype=np.float32)
        comb._load_images(l)

        logs = []
        lh = log_to_list(logger, logs, False)
        level = logger.getEffectiveLevel()
        logger.setLevel('DEBUG')

        if method == 'median':
            # for median, tot_size=5*4.5=22.5
            # xstep = 4, so n_chuks=25
            nchunk = 25
            shape = [(n, 4, 100)]
        elif method in ['mean', 'sum']:
            # for mean and sum, tot_size=5*3=15
            # xstep = 6, so n_chunks=16+1
            nchunk = 17
            shape = [(n, 6, 100), (n, 4, 100)]

        i = 0
        for chunk, unct, slc in comb._chunk_yielder(method=method):
            i += 1
            assert_in(chunk.shape, shape)
            assert_almost_equal(chunk[0], d[slc])
            assert_is_none(unct)
            assert_is_instance(chunk, np.ndarray)
        assert_equal(i, nchunk)
        assert_in(f'Splitting the images into {nchunk} chunks.', logs)
        logs.clear()

        # this should not split into chunks
        comb = ImCombiner(max_memory=1e8)
        comb._load_images(l)

        i = 0
        for chunk, unct, slc in comb._chunk_yielder(method=method):
            i += 1
            assert_equal(chunk.shape, (n, 100, 100))
            assert_equal(chunk[0], d)
            assert_is_none(unct)
            assert_is_instance(chunk, np.ndarray)
        assert_equal(i, 1)
        logs.clear()

    def test_chunk_yielder_ysplit(self):
        # using float32, the number of chunks are almost halved
        n = 100
        d = np.random.random((100, 100)).astype(np.float64)
        l = [FrameData(d, unit='adu') for i in range(n)]
        # data size = 4 000 000 = 4 bytes * 100 * 100 * 100
        # mask size = 1 000 000 = 1 bytes * 100 * 100 * 100
        # total size = 5 000 000

        logs = []
        lh = log_to_list(logger, logs, False)
        level = logger.getEffectiveLevel()
        logger.setLevel('DEBUG')

        # this should split in 300 chunks!
        # total_size = 4.5*5e6=22.5e6 = 225 chunks
        # x_step = 1
        # y_step = 45
        comb = ImCombiner(max_memory=1e5, dtype=np.float32)
        comb._load_images(l)
        i = 0
        for chunk, unct, slc in comb._chunk_yielder(method='median'):
            i += 1
            for k in chunk:
                assert_in(k.shape, ((1, 45), (1, 10)))
                assert_almost_equal(k, d[slc])
                assert_is_none(unct)
                assert_is_instance(k, np.ndarray)
        assert_equal(i, 300)
        assert_in('Splitting the images into 300 chunks.', logs)
        logs.clear()

        logger.setLevel(level)
        logger.removeHandler(lh)

    def test_chunk_yielder_uncertainty(self):
        n = 100
        d = np.random.random((100, 100)).astype(np.float64)
        u = np.random.random((100, 100)).astype(np.float64)
        l = [FrameData(d, uncertainty=u, unit='adu') for i in range(n)]

        # simple sum with uncertainties
        comb = ImCombiner(max_memory=2e6, dtype=np.float64)
        comb._load_images(l)
        i = 0
        for chunk, unct, slc in comb._chunk_yielder(method='sum'):
            i += 1
            for k,un in zip(chunk, unct):
                assert_in(k.shape, ((7, 100), (2, 100)))
                assert_almost_equal(k, d[slc])
                assert_almost_equal(un, u[slc])
                assert_is_instance(un, np.ndarray)
        assert_equal(i, 15)

        # if a single uncertainty is empty, disable it
        logs = []
        lh = log_to_list(logger, logs, False)
        level = logger.getEffectiveLevel()
        logger.setLevel('DEBUG')

        l[5].uncertainty = None
        comb = ImCombiner(max_memory=2e6, dtype=np.float64)
        comb._load_images(l)
        i = 0
        for chunk, unct, slc in comb._chunk_yielder(method='sum'):
            i += 1
            for k in chunk:
                assert_in(k.shape, ((7, 100), (2, 100)))
                assert_almost_equal(k, d[slc])
                assert_equal(unct, None)
        assert_equal(i, 15)
        assert_in('One or more frames have empty uncertainty. '
                  'Some features are disabled.',
                  logs)
        logs.clear()
        logger.setLevel(level)
        logger.removeHandler(lh)


class Test_ImCombiner_LoadImages():
    @pytest.mark.parametrize('disk_cache', [True, False])
    def test_image_loading_framedata(self, tmpdir, disk_cache):
        tmp = tmpdir.strpath
        n = 10
        d = np.ones((10, 10))
        l = [FrameData(d, unit='adu', uncertainty=d, cache_folder=tmp,
                       cache_filename=f'test{i}') for i in range(n)]

        comb = ImCombiner(use_disk_cache=disk_cache)
        # must start empty
        assert_equal(len(comb._images), 0)
        assert_is_none(comb._buffer)
        comb._load_images(l)
        assert_equal(len(comb._images), n)
        assert_is_none(comb._buffer)
        for i, v in enumerate(comb._images):
            fil = os.path.join(tmp, f'test{i}')
            assert_is_instance(v, FrameData)
            if disk_cache:
                assert_true(v._memmapping)
                # assert_path_exists(fil+'_copy_copy.data')
                # assert_path_exists(fil+'_copy_copy.unct')
                # assert_path_exists(fil+'_copy_copy.mask')

        comb._clear()
        # must start empty
        assert_equal(len(comb._images), 0)
        assert_is_none(comb._buffer)

        # ensure tmp files cleaned
        if disk_cache:
            for i in range(n):
                fil = os.path.join(tmp, f'test{i}')
                assert_path_not_exists(fil+'_copy_copy.data')
                assert_path_not_exists(fil+'_copy_copy.unct')
                assert_path_not_exists(fil+'_copy_copy.mask')

    @pytest.mark.parametrize('disk_cache', [True, False])
    def test_image_loading_fitsfile(self, tmpdir, disk_cache):
        tmp = tmpdir.strpath
        n = 10
        d = np.ones((10, 10))
        l = [os.path.join(tmp, f'fits_test{i}') for i in range(n)]
        for f in l:
            fits.PrimaryHDU(d).writeto(f)

        comb = ImCombiner(use_disk_cache=disk_cache)
        comb._load_images(l)

        assert_equal(len(comb._images), n)
        assert_is_none(comb._buffer)
        for i, v in enumerate(comb._images):
            assert_is_instance(v, FrameData)
            if disk_cache:
                assert_true(v._memmapping)

        comb._clear()
        # must start empty
        assert_equal(len(comb._images), 0)
        assert_is_none(comb._buffer)

    @pytest.mark.parametrize('disk_cache', [True, False])
    def test_image_loading_fitshdu(self, disk_cache):
        n = 10
        d = np.ones((10, 10))
        l = [fits.PrimaryHDU(d) for i in range(n)]

        comb = ImCombiner(use_disk_cache=disk_cache)
        comb._load_images(l)

        assert_equal(len(comb._images), n)
        assert_is_none(comb._buffer)
        for i, v in enumerate(comb._images):
            assert_is_instance(v, FrameData)
            if disk_cache:
                assert_true(v._memmapping)

        comb._clear()
        # must start empty
        assert_equal(len(comb._images), 0)
        assert_is_none(comb._buffer)

    def test_error_image_loading_empty(self):
        comb = ImCombiner()
        with pytest.raises(ValueError, match='Image list is empty.'):
            comb._load_images([])


class Test_ImCombiner_Rejection():
    @pytest.mark.parametrize('clip', [(0, 20), (0, None), (None, 20)])
    def test_apply_minmax_clip(self, clip):
        _min, _max = clip
        outliers = ((4, 2), (6, 2), (2, 9), (5, 2))
        outvalues = [1e6, -2e3, 5e3, -7e2]
        expect = np.zeros((10, 10))
        with NumpyRNGContext(123):
            data = np.random.normal(loc=10, scale=1, size=[10, 10])

        for p, v in zip(outliers, outvalues):
            data[p] = v
            if _min is not None:
                if _min > v:
                    expect[p] = 1
            if _max is not None:
                if _max < v:
                    expect[p] = 1

        data[0:2, 0:3] = np.nan
        expect[0:2, 0:3] = 1
        expect = expect.astype(bool)
        s = np.sum(expect)

        # force assign the buffer
        comb = ImCombiner()
        comb._buffer = data
        # with these limits, only the outliers must be masked
        comb.set_minmax_clip(_min, _max)
        comb._apply_rejection()
        # original mask must be kept
        sr = np.sum(np.isnan(comb._buffer))
        assert_equal(np.isnan(comb._buffer), expect)

    def test_apply_sigmaclip(self):
        data = np.array([1, -1, 1, -1, 65000], dtype=np.float32)
        comb = ImCombiner()

        # if threshold=1, only 65000 must be masked.
        comb._buffer = data.copy()
        comb.set_sigma_clip(1)
        expect = [0, 0, 0, 0, 1]
        comb._apply_rejection()
        assert_equal(np.isnan(comb._buffer), expect)

        # if threshold=3, 65000 must be masked too using mad_std.
        comb._buffer = data.copy()
        comb.set_sigma_clip(3)
        expect = [0, 0, 0, 0, 1]
        comb._apply_rejection()
        assert_equal(np.isnan(comb._buffer), expect)

        # if threshold=3 and a mask, the mask must be preserved.
        comb._buffer = data.copy()
        comb._buffer[1] = np.nan
        comb.set_sigma_clip(3)
        expect = [0, 1, 0, 0, 1]
        comb._apply_rejection()
        assert_equal(np.isnan(comb._buffer), expect)

    def test_apply_sigmaclip_only_lower(self):
        data = np.array([1, -1, 1, -1, 65000], dtype=np.float32)
        comb = ImCombiner()

        # 65000 must not be masked
        comb._buffer = data.copy()
        comb.set_sigma_clip((1, None))
        expect = [0, 0, 0, 0, 0]
        comb._apply_rejection()
        assert_equal(np.isnan(comb._buffer), expect)

    def test_apply_sigmaclip_only_higher(self):
        data = np.array([1, -1, 1, -1, -65000], dtype=np.float32)
        comb = ImCombiner()

        # -65000 must not be masked
        comb._buffer = data.copy()
        comb.set_sigma_clip((None, 1))
        expect = [0, 0, 0, 0, 0]
        comb._apply_rejection()
        assert_equal(np.isnan(comb._buffer), expect)


class Test_ImCombiner_Combine():
    @pytest.mark.parametrize('method', ['sum', 'median', 'mean'])
    def test_combine_mask_median(self, method):
        # Just check if mask combining works.
        images = [None]*10
        shape = (10, 10)
        for i in range(10):
            mask = np.zeros(shape)
            # all (2, 5) elements are masked, so the result must be
            mask[2, 5] = 1
            images[i] = FrameData(np.ones(shape), unit='adu', mask=mask)

        # these points in result must no be masked
        images[5].mask[7, 7] = 1
        images[2].mask[7, 7] = 1
        images[8].mask[8, 8] = 1
        expect = np.zeros(shape)
        expect[2, 5] = 1

        # with ImCombiner
        comb = ImCombiner()
        res = comb.combine(images, method)
        assert_equal(res.mask, expect)

        # with imcombine function
        assert_equal(imcombine(images, method).mask, expect)

    @pytest.mark.parametrize('method', ['sum', 'median', 'mean'])
    def test_combine_simple(self, method):
        values = [0.8, 1.0, 0.9, 1.1, 1.2, 1.1, 1.0, 0.8, 0.9]
        n = len(values)
        shape = (10, 10)
        images = [FrameData(np.ones(shape)*i, unit='adu') for i in values]
        if method == 'median':
            med = np.median(values)
            std = np.std(values)/np.sqrt(n)
        elif method == 'mean':
            med = np.mean(values)
            std = np.std(values)/np.sqrt(n)
        elif method == 'sum':
            med = np.sum(values)
            std = np.std(values)*np.sqrt(n)

        comb = ImCombiner()
        res1 = comb.combine(images, method=method)
        res2 = imcombine(images, method=method)
        for res in (res1, res2):
            assert_equal(res.data, np.ones(shape)*med)
            assert_almost_equal(res.uncertainty, np.ones(shape)*std)
            assert_equal(res.meta['astropop imcombine nimages'], n)
            assert_equal(res.meta['astropop imcombine method'], method)

    @pytest.mark.parametrize('method', ['sum', 'median', 'mean'])
    def test_combine_masked(self, method):
        values = [0.8, 1.0, 0.9, 1.1, 1.2, 1.1, 1.0, 0.8, 0.9]
        n = len(values)
        shape = (10, 10)
        images = [FrameData(np.ones(shape)*i, unit='adu') for i in values]

        # this pixel is masked and must be ignored
        images[0].mask[0, 0] = 1
        images[0].data[0, 0] = 1000

        if method == 'median':
            med = np.ones(shape)*np.median(values)
            std = np.ones(shape)*np.std(values)/np.sqrt(n)
            med[0, 0] = np.median(values[1:])
            std[0, 0] = np.std(values[1:])/np.sqrt(n-1)
        elif method == 'mean':
            med = np.ones(shape)*np.mean(values)
            std = np.ones(shape)*np.std(values)/np.sqrt(n)
            med[0, 0] = np.mean(values[1:])
            std[0, 0] = np.std(values[1:])/np.sqrt(n-1)
        elif method == 'sum':
            med = np.ones(shape)*np.sum(values)
            std = np.ones(shape)*np.std(values)*np.sqrt(n)
            med[0, 0] = np.sum(values[1:])*float(n)/(n-1)
            std[0, 0] = np.std(values[1:])*np.sqrt(n-1)*float(n)/(n-1)

        comb = ImCombiner()
        res1 = comb.combine(images, method=method)
        res2 = imcombine(images, method=method)
        for res in (res1, res2):
            assert_almost_equal(res.data, med)
            assert_almost_equal(res.uncertainty, std)
            assert_equal(res.meta['astropop imcombine nimages'], n)
            assert_equal(res.meta['astropop imcombine method'], method)



class Test_ImCombiner_HeaderMerging():
    def create_images(self):
        images = []
        for i in range(30):
            meta = {'first_equal': 1,
                    'second_equal': 2,
                    'first_differ': i,
                    'second_differ': i // 2,
                    'third_differ': i % 3}
            images.append(FrameData(np.ones((10, 10)), unit='adu', meta=meta))

        return images

    def test_mergeheaders_conformance(self):
        comb = ImCombiner()
        comb.set_merge_header('no_merge')
        comb.set_merge_header('first')
        comb.set_merge_header('only_equal')
        comb.set_merge_header('selected_keys', ['key 1'])
        with pytest.raises(ValueError, match='not known.'):
            comb.set_merge_header('unkown method')
        with pytest.raises(ValueError, match='No key assigned'):
            comb.set_merge_header('selected_keys')

    def test_mergeheaders_default(self):
        # default is 'no_merge'
        images = self.create_images()
        comb = ImCombiner()
        res = comb.combine(images, method='sum')
        for i, v in zip(['nimages', 'method'], [30, 'sum']):
            assert_equal(res.meta[f'astropop imcombine {i}'], v)

    def test_mergeheaders_no_merge(self):
        images = self.create_images()
        expect = {'astropop imcombine nimages': 30,
                  'astropop imcombine method': 'sum'}
        comb = ImCombiner(merge_header='no_merge')
        res = comb.combine(images, method='sum')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

        # explicit setting
        comb = ImCombiner()
        comb.set_merge_header('no_merge')
        res = comb.combine(images, method='sum')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

        # func
        res = imcombine(images, method='sum', merge_header='no_merge')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

    def test_mergeheaders_first(self):
        images = self.create_images()
        expect = images[0].meta
        expect.update({'astropop imcombine nimages': 30,
                       'astropop imcombine method': 'sum'})

        comb = ImCombiner(merge_header='first')
        res = comb.combine(images, method='sum')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

        # explicit setting
        comb = ImCombiner()
        comb.set_merge_header('first')
        res = comb.combine(images, method='sum')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

        # func
        res = imcombine(images, method='sum', merge_header='first')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

    def test_mergeheaders_only_equal(self):
        images = self.create_images()
        expect = {'astropop imcombine nimages': 30,
                  'astropop imcombine method': 'sum',
                  'first_equal': 1,
                  'second_equal': 2}

        comb = ImCombiner(merge_header='only_equal')
        res = comb.combine(images, method='sum')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

        # explicit setting
        comb = ImCombiner()
        comb.set_merge_header('only_equal')
        res = comb.combine(images, method='sum')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

        # func
        res = imcombine(images, method='sum', merge_header='only_equal')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

    def test_mergeheaders_selected_keys(self):
        images = self.create_images()
        keys = ['first_equal', 'third_differ', 'first_differ']
        expect = {'astropop imcombine nimages': 30,
                  'astropop imcombine method': 'sum',
                  'first_equal': 1,
                  'third_differ': 0,
                  'first_differ': 0}

        comb = ImCombiner(merge_header='selected_keys', merge_header_keys=keys)
        res = comb.combine(images, method='sum')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

        # explicit setting
        comb = ImCombiner()
        comb.set_merge_header('selected_keys', keys=keys)
        res = comb.combine(images, method='sum')
        for k, v in expect.items():
            assert_equal(res.meta[k], v)

        # func
        res = imcombine(images, method='sum', merge_header='selected_keys',
                        merge_header_keys=keys)
        for k, v in expect.items():
            assert_equal(res.meta[k], v)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Image proccessing library."""

from .imarith import imarith
from .imcombine import imcombine, ImCombiner


__all__ = ['imarith', 'imcombine', 'ImCombiner']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# flake8: noqa: F403, F405

import pytest
import numpy as np
from astropop.photometry.aperture import aperture_photometry, sky_annulus
from astropop.testing import *


@pytest.mark.parametrize('r', [2, 3, 4])
def test_simple_aperture(r):
    data = np.ones((10, 10))
    res = aperture_photometry(data, [5], [5], r=r, r_ann=None)
    assert_equal(res['x'], [5])
    assert_equal(res['y'], [5])
    assert_equal(res['aperture'], [r])
    assert_almost_equal(res['flux'], [np.pi*r**2])
    assert_not_in('sky', res.colnames)
    assert_almost_equal(res['flux_error'], [np.sqrt(np.pi*r**2)])
    assert_equal(res['flags'], [0])


data = np.array([[1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                 [1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                 [1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                 [1., 1., 1., 1., 2., 1., 1., 1., 1., 1.],
                 [1., 1., 1., 2., 4., 2., 1., 1., 1., 1.],
                 [1., 1., 1., 1., 2., 1., 1., 1., 1., 1.],
                 [1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                 [1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                 [1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                 [1., 1., 1., 1., 1., 1., 1., 1., 1., 1.]])

@pytest.mark.parametrize(['algoritmo','sky_kind'],
                         [['mmm', 'simple'],['sigmaclip','simple'],
                          ['mmm', '2and4'],['sigmaclip','2and4']])
def test_sky_annulus(algoritmo, sky_kind):

    if sky_kind == 'simple':
        # Simple homogeneous test:
        data = np.ones((20, 20))*3
        data[10,10] = 10
        data[9,9:12] = 5
        data[10,9] = 5
        data[10,11] = 5
        data[11,9:12] = 5

        sky_target = np.ones((1,8))*3
        sky_error_target = np.zeros((1,8))

    elif sky_kind == '2and4':
        # Heterogeneous test:
        data = np.ones((20, 20))*2
        data[1,:20] = 4
        data[3,:20] = 4
        data[5,:20] = 4
        data[7,:20] = 4
        data[1,:20] = 4
        data[9,:20] = 4
        data[11,:20] = 4
        data[13,:20] = 4
        data[15,:20] = 4
        data[17,:20] = 4
        data[19,:20] = 4
        data[10,10] = 10
        data[9,9:12] = 5
        data[10,9] = 5
        data[10,11] = 5
        data[11,9:12] = 5

        sky_target = np.ones((1,8))*3
        sky_error_target = np.ones((1,8))

    x = [9,9,10,10,10,11,11,11]
    y = [10,11,9,10,11,9,10,11]
    r_ann = [3, 5]

    sky, sky_error = sky_annulus(data, x, y, r_ann, algoritmo) #estudar como testar logger e mask

    assert_equal([sky], sky_target)
    assert_equal([sky_error], sky_error_target)



# phototils - verificar casos testes para fotometrias
# phototils
# background
# tests

# devdolarpix do IRAF








